# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#


"""Facilities for reading in files for testing."""

from pathlib import Path
from typing import Optional, Tuple, Union

from parselglossy.utils import JSONDict
from parselglossy.yaml_utils import read_yaml_file


def read_in(
    category: Union[str, Path], input_file_name: Optional[str], template_file_name: str
) -> Tuple[JSONDict, JSONDict]:
    """Helper function to read in input and validation template

    Parameters
    ----------
    category: str
    input_file_name: Optional[str]
    template_file_name: str

    Returns
    -------
    input_dict, template_dict: Tuple[JSONDict, JSONDict]
         A tuple with the two dictionaries.
    """
    this_path = Path(__file__).parent

    if input_file_name is None:
        input_dict = {}  # type: JSONDict
    else:
        input_file = this_path / "validation" / category / input_file_name
        input_dict = read_yaml_file(input_file)

    template_file = this_path / "validation" / category / template_file_name
    template_dict = read_yaml_file(template_file)

    return input_dict, template_dict



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Tests for `parselglossy` package.

Tests different views of validation specifications in YAML files.
"""

from contextlib import ExitStack as does_not_raise
from pathlib import Path

import pytest

from parselglossy import views
from parselglossy.yaml_utils import read_yaml_file

types = {
    "scf": {
        "functional": "str",
        "max_num_iterations": "int",
        "some_acceleration": "bool",
        "thresholds": {"some_integral_screening": "float", "energy": "float"},
        "another_number": "int",
        "some_complex_number": "complex",
    },
    "title": "str",
}

defaults = {
    "scf": {
        "another_number": 10,
        "functional": None,
        "max_num_iterations": 20,
        "some_acceleration": False,
        "some_complex_number": "0.0+0.0j",
        "thresholds": {"energy": 0.001, "some_integral_screening": 0.0001},
    },
    "title": None,
}

docstrings = {
    "scf": {
        "another_number": "Testing an even number in some range.",
        "functional": "XC functional.\n"
        "We also test that the string contains max 80 "
        "characters.",
        "max_num_iterations": "Max number of iterations.",
        "some_acceleration": "Turn on the new amazing acceleration " "scheme.",
        "some_complex_number": "Testing input with a complex number.",
        "thresholds": {
            "energy": "Some energy based threshold.",
            "some_integral_screening": "Some integral " "based " "threshold.",
        },
    },
    "title": "Title of the calculation.\n"
    "Since there is no default, this keyword is required.",
}

predicates = {
    "scf": {
        "another_number": [
            "0 <= value <= 40",
            "value % 2 == 0",
            "user['scf']['some_acceleration'] " "== True",
        ],
        "functional": ["len(value) <= 80"],
        "max_num_iterations": None,
        "some_acceleration": None,
        "some_complex_number": None,
        "thresholds": {"energy": None, "some_integral_screening": None},
    },
    "title": None,
}


@pytest.fixture
def template():
    this_path = Path(__file__).parent
    template_file = this_path / "validation" / "overall" / "template.yml"
    return read_yaml_file(template_file)


testdata = [
    (views.view_by_type, types, does_not_raise()),
    (views.view_by_default, defaults, does_not_raise()),
    (views.view_by_docstring, docstrings, does_not_raise()),
    (views.view_by_predicates, predicates, does_not_raise()),
    (lambda x: views.view_by("foobar", x), {}, pytest.raises(ValueError)),
]


@pytest.mark.parametrize(
    "viewer,reference,raises",
    testdata,
    ids=["types", "defaults", "docstrings", "predicates", "invalid"],
)
def test_view_by(template, viewer, reference, raises):
    with raises:
        view = viewer(template)
        assert view == reference



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Atoms."""

import functools
from typing import Any, List, Union

try:
    import pyparsing as pp
except ImportError:
    # Import local copy
    from . import pyparsing as pp  # type: ignore


TRUTHY = ["TRUE", "ON", "YES", "Y"]
"""List[str]: List of true-like values."""

FALSEY = ["FALSE", "OFF", "NO", "N"]
"""List[str]: List of false-like values."""


def to_bool(x):
    defined = False
    if x is None:
        defined = False
    elif x.upper() in FALSEY:
        defined = False
    elif x.upper() in TRUTHY:
        defined = True
    else:
        defined = False

    return defined


bool_t = functools.reduce(lambda x, y: x ^ y, map(pp.CaselessLiteral, TRUTHY + FALSEY))
bool_t.setName("bool")
bool_t.setParseAction(lambda token: to_bool(token[0]))

int_t = pp.pyparsing_common.signed_integer

float_t = pp.pyparsing_common.sci_real

quoted_str_t = pp.quotedString.setParseAction(pp.removeQuotes)
unquoted_str_t = pp.Word(pp.alphas + "_", pp.alphanums + "_")
"""An unquoted string starts with alphabetic characters and underscores,
followed by alphanumeric characters and underscores."""

I_unit = functools.reduce(
    lambda x, y: x ^ y, map(pp.CaselessLiteral, ["*j", "*i"])
).suppress()
complex_t = pp.OneOrMore(pp.pyparsing_common.number) + I_unit
complex_t.setParseAction(
    lambda token: complex(token[0], token[1])
    if len(token) == 2
    else complex(0.0, token[0])
)

num_t = complex_t | float_t | int_t
num_t.setName("numeric")

SDATA = pp.Literal("$").suppress()
EDATA = pp.CaselessLiteral("$end").suppress()
data_t = pp.Group(
    pp.Combine(SDATA + pp.Word(pp.alphas + "_<>", pp.alphanums + "_<>"))
    + pp.SkipTo(EDATA)
    + EDATA
)

fortranStyleComment = pp.Regex(r"!.*").setName("Fortran style comment")


def make_list_t(
    scalars: Union[Any, List[Any]],
    *,
    start: str = "[",
    end: str = "]",
    delimiter: str = ",",
    throw_if_empty: bool = True,
    multiline: bool = True
) -> Any:
    """Atom for lists.

    Parameters
    ----------
    scalars: Union[Any, List[Any]]
        Scalar parser elements, already combined or as a list. The list will be
        combined using the `^` operator.
    start: str
        Left delimiter for the list. Defaults to '['.
    end: str
        Right delimiter for the list. Defaults to ']'.
    delimiter: str
        List delimiter. Defaults to ','.
    multiline: bool
        Whether the list can span multiple lines. Defaults to `True`.

    Notes
    -----
    The order of the scalar tokens in the `scalars` list **is important**.

    Returns
    -------
    list: Any
    """
    START, END = map(pp.Suppress, start + end)

    if isinstance(scalars, list):
        atoms = functools.reduce(lambda x, y: x ^ y, scalars)
    else:
        atoms = scalars

    if multiline:
        NEWLINE = pp.Literal("\n").suppress()
        list_t = (
            START
            + pp.Optional(pp.delimitedList(atoms ^ NEWLINE, delim=delimiter))
            + END
        )
    else:
        list_t = START + pp.Optional(pp.delimitedList(atoms, delim=delimiter)) + END

    return list_t.addCondition(
        bool, message="Empty lists not allowed", fatal=throw_if_empty
    )


list_t = make_list_t(quoted_str_t ^ float_t ^ int_t ^ unquoted_str_t ^ bool_t)



# -*- coding: utf-8 -*-

import pathlib
import sys

import guzzle_sphinx_theme
from sphinx.ext.apidoc import main

sys.path.insert(0, str(pathlib.Path(__file__).parents[1]))

from parselglossy import __version__ as _version  # isort:skip
from parselglossy import __author__ as _author  # isort:skip

# -- General configuration ---------------------------------------------

extensions = ["sphinx.ext.napoleon", "sphinx.ext.viewcode", "guzzle_sphinx_theme"]
templates_path = ["_templates"]
source_suffix = ".rst"
master_doc = "index"
project = u"parselglossy"
copyright = u"2018, dev-cafe"
author = _author
# The short X.Y version.
version = _version
# The full version, including alpha/beta/rc tags.
release = _version
language = None
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]
pygments_style = "sphinx"
todo_include_todos = False

# -- Options for HTML output -------------------------------------------
html_title = "parselglossy Documentation"
html_short_title = f"parselglossy {version}"
html_show_sourcelink = False
html_sidebars = {
    "**": ["logo-text.html", "globaltoc.html", "localtoc.html", "searchbox.html"]
}
html_theme_path = guzzle_sphinx_theme.html_theme_path()
html_theme = "guzzle_sphinx_theme"
html_theme_options = {
    # Set the name of the project to appear in the sidebar
    "project_nav_name": "parselglossy"
}


def run_apidoc(_):
    cur_dir = pathlib.Path(__file__).parent
    module = cur_dir.parent / project
    main(["-e", "-o", str(cur_dir), str(module), "--force"])


def setup(app):
    app.connect("builder-inited", run_apidoc)



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

import re
from typing import Any, Callable, List, Optional, Tuple, Union

from typing import Dict  # noqa: F401; noqa: F401

RunCallable = Callable[[Any, str], Tuple[str, Any]]

ScalarTypes = Union[bool, str, int, float, complex]

allowed_scalar_types = ["str", "int", "float", "complex", "bool"]


ListTypes = Union[List[bool], List[str], List[int], List[float], List[complex]]

allowed_list_types = [f"List[{t}]" for t in allowed_scalar_types]

AllowedTypes = Union[ScalarTypes, ListTypes]

allowed_types = allowed_scalar_types + allowed_list_types


def type_matches(value: AllowedTypes, expected_type: str) -> Optional[bool]:
    """Checks whether a value is of the expected type.

    Parameters
    ----------
    value : AllowedTypes
        Value whose type needs to be checked
    expected_type : str

    Notes
    -----
    Allowed types T are: `str`, `int`, `float`, `complex`, `bool`,
    as well as `List[T]`.

    Returns
    -------
    True if value has the type expected_type, otherwise False.

    Raises
    ------
    ValueError
        If expected_type is not among the allowed types.

    Notes
    -----
    Complex numbers are a tad more fastidious, as they *might be* read in as
    strings. To avoid false negatives, we have a nasty hack, for which RDR will
    forever burn in hell. For complex and List[complex] we will re-run type
    checking after type
    casting if and only if:
      - We expected complex and List[complex].
      - We got a str or List[str].
      - The string are in the right format for the type casting operator.
    """

    # first verify whether expected_type is allowed
    if expected_type not in allowed_types:
        raise ValueError(f"could not recognize expected_type: {expected_type}")

    expected_complex = expected_type == "complex" or expected_type == "List[complex]"

    expected_type_is_list = re.search(r"^List\[(\w+)\]$", expected_type)

    matches = False
    if expected_type_is_list is not None:
        matches = _type_check_list(
            value, expected_type_is_list.group(1)
        )  # type: ignore
    else:
        matches = _type_check_scalar(value, expected_type)  # type: ignore

    recheck_complex = (
        expected_complex
        and _scalar_type_is_str(value)
        and _str_is_complex_number(value)
    )
    if not matches and recheck_complex:
        # Run the type cast, this might fail with ValueError and return None.
        value = _complex_helper(value)  # type: ignore
        if value is not None:
            matches = type_matches(value, expected_type)  # type:ignore
        else:
            matches = False

    return matches


def _complex_helper(
    value: Union[complex, List[complex]]
) -> Optional[Union[complex, List[complex]]]:
    try:
        if type(value).__name__ == "list":
            return type_fixers["List[complex]"](value)
        else:
            return type_fixers["complex"](value)
    except ValueError:
        return None


def _scalar_type_is_str(value: AllowedTypes) -> bool:
    return any(
        [_type_check_scalar(value, "str"), _type_check_list(value, "str")]
    )  # type: ignore


def _str_is_complex_number(value: AllowedTypes) -> bool:
    cmplx = re.compile(r"^.*\d+j$")
    if type(value).__name__ == "list":
        return all((cmplx.match(x) is not None for x in value))
    else:
        return cmplx.match(value) is not None


def _type_check_scalar(value: ScalarTypes, expected_type: str) -> bool:
    return type(value).__name__ == expected_type


def _type_check_list(value: ListTypes, expected_type: str) -> bool:
    if type(value).__name__ == "list":
        type_checks = all((_type_check_scalar(x, expected_type) for x in value))
    else:
        type_checks = False

    return type_checks


type_fixers = {
    "bool": bool,
    "complex": complex,
    "float": float,
    "int": int,
    "str": str,
    "List[bool]": lambda x: list(map(bool, x)),
    "List[complex]": lambda x: list(map(complex, x)),
    "List[float]": lambda x: list(map(float, x)),
    "List[int]": lambda x: list(map(int, x)),
    "List[str]": lambda x: list(map(str, x)),
}  # type: Dict[str, Callable[[Any], Any]]
"""Dict[str, Callable[[Any], Any]]: dictionary holding functions for type fixation."""



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Plumbing functions powering our validation facilities."""

from copy import deepcopy
from typing import Any, List, Optional, Tuple

from .exceptions import Error
from .types import type_fixers, type_matches
from .utils import JSONDict, location_in_dict, nested_set


def _rec_merge_ours(
    *, theirs: JSONDict, ours: JSONDict, address: Tuple = ()
) -> Tuple[JSONDict, List[Error]]:
    """Recursively merge two ``dict``-s with "ours" strategy.

    Parameters
    ----------
    theirs : JSONDict
    ours : JSONDict
    address : Tuple

    Returns
    -------
    outgoing : JSONDict
    errors : List[Error]

    Notes
    -----
    The ``theirs`` dictionary is supposed to be the view by defaults of the
    validation specification, whereas ``ours`` is the dictionary from user
    input. The recursive merge action will generate a complete, but not
    validated, input dictionary by using default values where these are not
    overridden by user input, hence the naming "ours" for the merge strategy.
    """
    outgoing = {}
    errors = []

    # Check whether ours has keywords/sections that are unknown
    difference = set(ours.keys()).difference(set(theirs.keys()))
    if difference != set():
        for k in difference:
            what = "section" if isinstance(ours[k], dict) else "keyword"
            errors.append(Error(message=f"Found unexpected {what}: '{k}'."))

    for k, v in theirs.items():
        if k not in ours.keys():
            if theirs[k] is not None:
                outgoing[k] = theirs[k]
            else:
                outgoing[k] = None
                msg = f"Keyword '{k}' is required but has no value."
                errors.append(Error((address + (k,)), msg))
        elif not isinstance(v, dict):
            outgoing[k] = ours[k]
        else:
            outgoing[k], errs = _rec_merge_ours(
                theirs=v, ours=ours[k], address=(address + (k,))
            )
            errors.extend(errs)

    return outgoing, errors


def _rec_fix_defaults(
    incoming: JSONDict,
    *,
    types: JSONDict,
    start_dict: JSONDict = None,
    address: Tuple = (),
) -> Tuple[JSONDict, List[Error]]:
    """Fix default values and perform type checking.

    Parameters
    ----------
    incoming : JSONDict
        The input ``dict``. This is supposed to be the one obtained by merging
        user and template ``dict``-s.
    types: JSONDict
        Types of all keywords in the input. Generated from :func:`view_by_types`.
    start_dict : JSONDict
        The ``dict`` we start recursion from. This parameter is needed to keep
        around a copy of the full ``dict`` during the recursion.
    address : Tuple[str]
        A tuple of keys need to index the current value in the recursion. See
        Notes.

    Returns
    -------
    outgoing : JSONDict
        A dictionary with all default values fixed.
    errors : List[Error]
        A list of keys to access elements in the `dict` that raised an error.
        See Notes.

    Notes
    -----
    Since we allow callables to appear as defaults, we need to run them
    to determine the actual default values.

    This operation must be done with some care, to avoid false negatives or
    ambiguous type checks. For example:

    * If the type is ``str`` and the default a callable, the type will
      match, but the default will make no sense.
    * If the type is numerical, *e.g.* ``int``, the type will not match.

    However, by design the callables **must** refer to some other field in the
    input tree, hence they **must** contain the reserved token "user". This
    allows us to disambiguate a callable as default from a value as default.

    The final strategy adopted is then:

    1. Perform type checking with :func:`type_matches`. If successful, we
    coerce the type.
    2. If types did not match, we further check whether the value is a string,
    containing the reserved token "value". This means the default value is
    actually a callable. We run the callable, which internally coerces the type
    of the result to the expected one.
    3. If even this check was unsuccessful, types really were unmatched. We
    report the error and move on.
    """

    if start_dict is None:
        start_dict = deepcopy(incoming)

    outgoing = {}
    errors = []

    for k, v in incoming.items():
        if not isinstance(v, dict):
            t = types[k]
            types_ok = type_matches(v, t)
            if types_ok:
                # Yes! Types match up front, you're awesome
                msg = ""
                outgoing[k] = type_fixers[t](v)
            else:
                # Types did not match :/
                if isinstance(v, str) and "user" in v:
                    # BUT! It's actually a string and it's a callable
                    # We assume that if it contains the reserved tokens "value"
                    # or "user" we're trying to perform some sort of defaulting
                    # action.
                    msg, outgoing[k] = run_callable(v, start_dict, t=t)
                else:
                    # NOPE. You're an unrepentant sinner
                    actual = (
                        type(v).__name__
                        if type(v) is not list
                        else f"List[{', '.join([type(x).__name__ for x in v])}]"
                    )
                    msg = f"Actual ({actual}) and declared ({t}) types do not match."
            if msg != "":
                errors.append(Error(address + (k,), msg))
            else:
                # Update start_dict.
                # This is so that multiple dependent defaults ("chains") behave
                # correctly. See #76 on GitHub
                nested_set(start_dict, address + (k,), outgoing[k])
        else:
            outgoing[k], errs = _rec_fix_defaults(
                incoming=v,
                types=types[k],
                start_dict=start_dict,
                address=(address + (k,)),
            )
            errors.extend(errs)

    return outgoing, errors


def _rec_check_predicates(
    incoming: JSONDict,
    *,
    predicates: JSONDict,
    start_dict: JSONDict = None,
    address: Tuple = (),
) -> List[Error]:
    """Run predicates on input tree with fixed defaults.

    Parameters
    ----------
    incoming : JSONDict
        The input `dict`. This is supposed to be the result of :func:`fix_defaults`.
    predicates : JSONDict
        A view-by-predicates of the template ``dict``.
    start_dict : JSONDict
        The `dict` we start recursion from.
    address : Tuple[str]
        A tuple of keys need to index the current value in the recursion.

    Returns
    -------
    errors : List[Error]
        A list of keys to access elements in the `dict` that raised an error.
    """

    errors = []

    if start_dict is None:
        start_dict = incoming

    for k, v in incoming.items():
        if predicates[k] is not None:
            if not isinstance(v, dict):
                for p in predicates[k]:
                    where = location_in_dict(address=(address + (k,)))
                    msg, success = run_predicate(p, where, start_dict)
                    if not success:
                        errors.append(Error((address + (k,)), msg))
            else:
                errs = _rec_check_predicates(
                    incoming=v,
                    predicates=predicates[k],
                    start_dict=start_dict,
                    address=(address + (k,)),
                )
                errors.extend(errs)

    return errors


def run_predicate(predicate: str, where: str, user: JSONDict) -> Tuple[str, bool]:
    """Run a predicate to check whether it is satisfied.

    Parameters
    ----------
    predicate : str
    where : str
    user : JSONDict

    Returns
    -------

    Notes
    -----
    We replace the convenience placeholder "value" with its full "address" in ``user``.
    """

    p = predicate.replace("value", where)

    try:
        msg = ""
        success = eval(f"lambda user: {p}")(user)
        if not success:
            msg = f"Predicate '{predicate}' not satisfied."
    except KeyError as e:
        msg = f"KeyError {e} in closure '{predicate}'."
        success = False
    except SyntaxError as e:
        msg = f"SyntaxError {e} in closure '{predicate}'."
        success = False
    except TypeError as e:
        msg = f"TypeError {e} in closure '{predicate}'."
        success = False
    except NameError as e:
        msg = f"NameError {e} in closure '{predicate}'."
        success = False

    return msg, success


def run_callable(f: str, d: JSONDict, *, t: str) -> Tuple[str, Optional[Any]]:
    """Run a callable encoded as a string.

    A callable is any function of the input tree.

    Parameters
    ----------
    f : str
        Callable to checked as a string
    d : JSONDict
        The input `dict`.
    t : str
        Expected type.

    Returns
    -------
    retval : Tuple[str, Optional[Any]]
        The error message, if any, and the result of the callable, if any.

    Notes
    -----
    The input tree is called ``user``.
    The callable is turned into a lambda function and executed using ``eval``,
    to ensure that the syntax of callable actions is correct and that the
    callable returns correctly.

    We need to pass the full ``incoming`` dictionary as argument to
    ``eval``, because we allow indexing in the *global* ``dict``. That is,
    since it is **allowed** to define defaults in a given section based on
    defaults in other section we **must** be able to access the full input at
    any point.
    """

    try:
        msg = ""
        result = eval(f"lambda user: {f}")(d)
        result = type_fixers[t](result)
    except KeyError as e:
        msg = f"KeyError {e} in closure '{f}'."
        result = None
    except SyntaxError as e:
        msg = f"SyntaxError {e} in closure '{f}'."
        result = None
    except TypeError as e:
        msg = f"TypeError {e} in closure '{f}'."
        result = None
    except NameError as e:
        msg = f"NameError {e} in closure '{f}'."
        result = None

    return msg, result



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Tests for `parselglossy` package.

Tests checking predicates.
"""

import re
from contextlib import ExitStack as does_not_raise
from copy import deepcopy

import pytest

from parselglossy.exceptions import ParselglossyError
from parselglossy.validation import check_predicates


@pytest.fixture
def user():
    return {
        "scf": {
            "another_number": 10,
            "functional": "B3LYP",
            "max_num_iterations": 20,
            "some_acceleration": False,
            "some_complex_number": complex("0.0+0.0j"),
            "thresholds": {"energy": 0.001, "some_integral_screening": 0.0001},
        },
        "title": "My title",
    }


empty_predicates = {
    "scf": {
        "another_number": None,
        "functional": None,
        "max_num_iterations": None,
        "some_acceleration": None,
        "some_complex_number": None,
        "thresholds": {"energy": None, "some_integral_screening": None},
    },
    "title": None,
}


def valid_predicates():
    d = deepcopy(empty_predicates)
    d["scf"]["another_number"] = ["value < user['scf']['max_num_iterations']"]
    return d


def failing_predicates():
    d = deepcopy(empty_predicates)
    d["scf"]["another_number"] = ["value % 3 == 0"]
    return d


def syntax_error_predicates():
    d = deepcopy(empty_predicates)
    d["scf"]["another_number"] = ["value =< user['scf']['max_num_iterations']"]
    return d


def name_error_predicates():
    d = deepcopy(empty_predicates)
    d["scf"]["another_number"] = ["foo <= value < user['scf']['max_num_iterations']"]
    return d


def key_error_predicates():
    d = deepcopy(empty_predicates)
    d["scf"]["another_number"] = ["value < user['scf']['min_num_iterations']"]
    return d


def type_error_predicates():
    d = deepcopy(empty_predicates)
    d["scf"]["another_number"] = ["value < user['scf']['max_num_iterations'] / 'foooo'"]
    return d


def multiple_errors_predicates():
    d = deepcopy(empty_predicates)
    d["title"] = ["len(value) < user['foo']"]
    d["scf"]["another_number"] = [
        "value / 'fooo' < 10",
        "value < user['scf']['max_num_iterations']",
    ]
    return d


testdata = [
    (valid_predicates(), None, [""]),
    (
        failing_predicates(),
        pytest.raises(ParselglossyError),
        [
            r"Error(?:s)? occurred when checking predicates:\n- At user\['scf'\]\['another_number'\]:\s+Predicate 'value % 3 == 0' not satisfied\."
        ],
    ),
    (
        syntax_error_predicates(),
        pytest.raises(ParselglossyError),
        [
            r"Error(?:s)? occurred when checking predicates:\n- At user\['scf'\]\['another_number'\]:\s+SyntaxError.*"
        ],
    ),
    (
        name_error_predicates(),
        pytest.raises(ParselglossyError),
        [
            r"Error(?:s)? occurred when checking predicates:\n- At user\['scf'\]\['another_number'\]:\s+NameError.*"
        ],
    ),
    (
        key_error_predicates(),
        pytest.raises(ParselglossyError),
        [
            r"Error(?:s)? occurred when checking predicates:\n- At user\['scf'\]\['another_number'\]:\s+KeyError.*"
        ],
    ),
    (
        type_error_predicates(),
        pytest.raises(ParselglossyError),
        [
            r"Error(?:s)? occurred when checking predicates:\n- At user\['scf'\]\['another_number'\]:\s+TypeError.*"
        ],
    ),
    (
        multiple_errors_predicates(),
        pytest.raises(ParselglossyError),
        [
            r"Error(?:s)? occurred when checking predicates:\n- At user\['scf'\]\['another_number'\]:\s+TypeError.*\n- At user\['title'\]:\s+KeyError.*",
            r"Error(?:s)? occurred when checking predicates:\n- At user\['title'\]:\s+KeyError.*\n- At user\['scf'\]\['another_number'\]:\s+TypeError.*",
        ],
    ),
]


@pytest.mark.parametrize(
    "predicates,raises,error_message",
    testdata,
    ids=[
        "valid",
        "failing",
        "syntax_error",
        "name_error",
        "key_error",
        "type_error",
        "two_errors",
    ],
)
def test_check_predicates(user, predicates, raises, error_message):
    if raises is None:
        ctx = does_not_raise()
    else:
        ctx = pytest.raises(ParselglossyError, match="|".join(error_message))

    with ctx:
        check_predicates(incoming=user, predicates=predicates)



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Common utilities."""

import json
from functools import reduce
from pathlib import Path
from shutil import copy
from typing import Any, Dict, List, Optional, Tuple, Union

JSONDict = Dict[str, Any]


class ComplexEncoder(json.JSONEncoder):
    """JSON encoder for complex numbers."""

    def default(self, obj):
        if isinstance(obj, complex):
            return {"__complex__": [obj.real, obj.imag]}
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)


def as_complex(dct):
    """JSON decoder for complex numbers."""
    if "__complex__" in dct:
        return complex(dct["__complex__"][0], dct["__complex__"][1])
    return dct


def location_in_dict(*, address: Tuple, dict_name: str = "user") -> str:
    """Convert tuple of keys of a ``JSONDict`` to its representation in code.

    For example, given ``("a", "b", "c")`` returns the string ``user['a']['b']['c']``.

    Parameters
    ----------
    address : Tuple[str]
    dict_name : str

    Returns
    -------
    where : str
    """
    return reduce(lambda x, y: x + f"['{y}']", address, "user")


def path_resolver(f: Union[str, Path], *, touch: bool = True) -> Path:
    """Resolve a path.

    Parameters
    ----------
    f : Union[str, Path]
        File whose path needs to be resolved.
    touch : bool
        Create file is not already existent. Default: True

    Returns
    -------
    path : Path
        File as a ``Path`` object.

    Notes
    -----
    The file will be created if not already existent.
    """

    path = Path(f) if isinstance(f, str) else f

    if not path.exists() and touch:
        path.touch()

    return path.resolve()


def default_outfile(*, fname: Union[str, Path], suffix: str) -> str:
    """Default name for output file.

    Parameters
    ----------
    fname : Union[str, Path]
        Name to use as stencil.
    suffix : str
        Suffix to append.

    Returns
    -------
    The name of the output file.
    """
    fname = Path(fname) if isinstance(fname, str) else fname

    base = fname.name

    return base.rsplit(".", 1)[0] + suffix


def copier(src: Path, dest: Path) -> None:
    """Copy file, ensuring it can be overwritten."""
    # copy file
    _ = copy(src, dest, follow_symlinks=True)
    # Ensure we can overwrite, by doing a chmod uga+rw
    Path(_).chmod(0o666)


def flatten_list(nested_list: List[Any]) -> List[Any]:
    """Flattens a nested list into a flat list.

    Parameters
    ----------
    nested_list : List[Any]
         Nested list

    Returns
    -------
    List[Any]
         Flattened list
    """
    if nested_list == []:
        return nested_list
    if isinstance(nested_list[0], list):
        return flatten_list(nested_list[0]) + flatten_list(nested_list[1:])
    return nested_list[:1] + flatten_list(nested_list[1:])


def dict_to_list(d: JSONDict) -> List[Any]:
    """Converts a nested dictionary to a nested list.

    Parameters
    ----------
    d : JSONDict
         Nested dictionary

    Returns
    -------
    list: List[Any]
         Flattened list
    """
    nested_list = []
    for k, v in d.items():
        if isinstance(v, dict):
            nested_list.append([k, dict_to_list(v)])
        else:
            nested_list.append([k, v])
    return nested_list


def nested_get(d: JSONDict, *ks: str) -> Optional[Any]:
    """Get value from a nested dictionary.

    Parameters
    ----------
    d : JSONDict
    ks : str

    Returns
    -------
    v : Optional[Any]

    Notes
    -----
    Adapted from: https://stackoverflow.com/a/40675868/2528668
    """

    def _func(x: JSONDict, k: str) -> Optional[JSONDict]:
        return x.get(k, None) if isinstance(x, dict) else None

    return reduce(_func, ks, d)  # type: ignore


def nested_set(d: JSONDict, ks: Tuple[Any, ...], v: Any) -> None:
    """Set value in nested dictionary.

    Parameters
    ----------
    d : JSONDict
    ks : Tuple[str]
    v : Any

    Notes
    -----
    Adapted from: https://stackoverflow.com/a/13688108/2528668
    """
    for k in ks[:-1]:
        d = d.setdefault(k, {})
    d[ks[-1]] = v



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Tests for `parselglossy` package."""

from string import ascii_letters, digits

import pytest
from hypothesis import example, given
from hypothesis import strategies as st
from pyparsing import ParseBaseException

from custom_strategies import (
    complex_numbers,
    floats,
    list_of_complex_numbers,
    list_of_floats,
    list_of_unquoted_str,
    unquoted_str,
)
from parselglossy.grammars import atoms

PRINTABLE = ascii_letters + digits + r"!#$%&*+-./:;<>?@^_|~"
"""str: Custom printable character set.

The printable character set is the standard set in `string.printable` minus
"\'(),=[\\]`{} and all whitespace characters.
"""


@given(a=st.sampled_from(atoms.TRUTHY + atoms.FALSEY))
def test_atom_bool(a):
    tokens = atoms.bool_t.parseString(f"{a:s}").asList()
    assert tokens[0] == atoms.to_bool(a)


@given(a=st.integers())
def test_atoms_int(a):
    tokens = atoms.int_t.parseString(f"{a:d}").asList()
    assert tokens[0] == a


@given(a=floats())
def test_atoms_float(a):
    tokens = atoms.float_t.parseString(a.string).asList()
    assert tokens[0] == pytest.approx(a.value)


@given(a=unquoted_str())
@example(a="foo_BAR")
@example(a="_123rdfoo_BAR")
def test_atoms_unquoted_str(a):
    tokens = atoms.unquoted_str_t.parseString(a).asList()
    assert tokens[0] == a


@pytest.mark.parametrize(
    "quoting", ["'{:s}'", '"{:s}"'], ids=["single_quotes", "double_quotes"]
)
@given(a=st.text(alphabet=(PRINTABLE + " "), min_size=1))
@example(a="Bobson Dugnutt")
@example(a="Glenallen    Mixon   ")
@example(a="    Todd Bonzalez   ")
@example(a="Dwigt Rortugal")
@example(a="Raul_Chamgerlain    ")
def test_atoms_quoted_str(a, quoting):
    tokens = atoms.quoted_str_t.parseString(quoting.format(a)).asList()
    assert tokens[0] == a


@given(a=complex_numbers())
def test_atoms_complex(a):
    tokens = atoms.complex_t.parseString(a.string).asList()
    assert tokens[0] == pytest.approx(a.value)


@given(
    a=st.sampled_from(
        [
            ["raw", "H 0.0 0.0 0.0\nF 1.0 1.0 1.0\n"],
            ["RAW", "H 0.0 0.0 0.0\nF 1.0 1.0 1.0\n"],
        ]
    ).map(lambda x: ("${0:s}\n{1:s}$end".format(x[0], x[1]), x))
)
def test_atoms_data(a):
    tokens = atoms.data_t.parseString(a[0]).asList()
    assert tokens[0] == a[1]


@pytest.mark.parametrize("a", ["[]", "[       ]", "[\n]"])
def test_list_empty(a):
    with pytest.raises(ParseBaseException, match=r"^Empty lists not allowed"):
        _ = atoms.list_t.parseString(f"{a}").asList()


@given(a=st.lists(st.integers(), min_size=1))
def test_list_int(a):
    tokens = atoms.list_t.parseString(f"{a}").asList()
    assert tokens == a


@given(a=list_of_floats())
def test_list_float(a):
    tokens = atoms.list_t.parseString(a.string).asList()
    assert tokens == pytest.approx(a.value)


@given(a=list_of_unquoted_str())
def test_list_unquoted_str(a):
    tokens = atoms.list_t.parseString(a.string).asList()
    assert tokens == a.value


@pytest.mark.parametrize(
    "quoting", ["'{:s}'", '"{:s}"'], ids=["single_quotes", "double_quotes"]
)
@given(a=st.lists(st.text(alphabet=(PRINTABLE)), min_size=1))
def test_list_quoted_str(a, quoting):
    example = "[" + ", ".join([quoting.format(x) for x in a]) + "]"
    tokens = atoms.list_t.parseString(example).asList()
    assert tokens == a


@given(a=list_of_complex_numbers())
def test_list_complex(a):
    scalar = (
        atoms.quoted_str_t
        ^ atoms.complex_t
        ^ atoms.float_t
        ^ atoms.int_t
        ^ atoms.unquoted_str_t
        ^ atoms.bool_t
    )
    list_t = atoms.make_list_t(scalar)
    tokens = list_t.parseString(a.string).asList()
    assert tokens == pytest.approx(a.value)



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

from pathlib import Path
from typing import Union

import yaml

from .utils import JSONDict


def read_yaml_file(file_name: Union[str, Path]) -> JSONDict:
    """Reads a YAML file and returns it as a dictionary.

    Parameters
    ----------
    file_name: Union[str, Path]
        Path object for the YAML file.

    Returns
    -------
    d: JSONDict
        A dictionary with the contents of the YAML file.
    """

    file_name = Path(file_name) if isinstance(file_name, str) else file_name

    with file_name.open("r") as f:
        try:
            d = yaml.safe_load(f)
        except yaml.YAMLError as e:
            print(e)
    return d



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Tests for `parselglossy` package."""

import pytest
from hypothesis import given
from hypothesis import strategies as st

from parselglossy.types import type_matches


@given(a=st.booleans())
def test_type_matches_bool(a):
    assert type_matches(a, "bool")
    assert not type_matches(0, "bool")


@given(a=st.text())
def test_type_matches_str(a):
    assert type_matches(a, "str")
    assert not type_matches(a, "int")
    assert not type_matches(a, "float")
    assert not type_matches(a, "complex")
    assert not type_matches(1.0, "str")


@given(a=st.integers())
def test_type_matches_int(a):
    assert type_matches(a, "int")
    assert not type_matches(a, "str")
    assert not type_matches(a, "float")
    assert not type_matches(a, "complex")


@given(a=st.floats(allow_nan=False, allow_infinity=False))
def test_type_matches_float(a):
    assert type_matches(a, "float")
    assert not type_matches(a, "str")
    assert not type_matches(a, "int")
    assert not type_matches(a, "complex")
    assert not type_matches(1, "float")


@given(a=st.complex_numbers(allow_nan=False, allow_infinity=False))
def test_type_matches_complex(a):
    assert type_matches(a, "complex")
    assert not type_matches(a, "str")
    assert not type_matches(a, "int")
    assert not type_matches(a, "float")


@given(a=st.lists(st.text(), min_size=1))
def test_type_matches_list_str(a):
    assert type_matches(a, "List[str]")
    assert not type_matches((1, 2, 3), "List[str]")
    assert not type_matches([1, 2, 3], "List[str]")
    assert not type_matches(["foo", 2, 3], "List[str]")
    assert not type_matches(a, "List[int]")
    assert not type_matches(a, "List[float]")
    assert not type_matches(a, "List[complex]")


@given(a=st.lists(st.integers(), min_size=1))
def test_type_matches_list_int(a):
    assert type_matches(a, "List[int]")
    assert not type_matches((1, 2, 3), "List[int]")
    assert not type_matches([1, "foo", 2, 3], "List[int]")
    assert not type_matches(a, "List[str]")
    assert not type_matches(a, "List[float]")
    assert not type_matches(a, "List[complex]")


@given(a=st.lists(st.floats(allow_nan=False, allow_infinity=False), min_size=1))
def test_type_matches_list_float(a):
    assert type_matches(a, "List[float]")
    assert not type_matches((1, 2, 3), "List[float]")
    assert not type_matches([1, 2, 3], "List[float]")
    assert not type_matches([1.0, 2, 3], "List[float]")
    assert not type_matches(a, "List[str]")
    assert not type_matches(a, "List[int]")
    assert not type_matches(a, "List[complex]")


@given(
    a=st.lists(st.complex_numbers(allow_nan=False, allow_infinity=False), min_size=1)
)
def test_type_matches_list_complex(a):
    assert type_matches(a, "List[complex]")
    assert not type_matches((1 + 1j, 2 + 2j, 3 + 3j), "List[complex]")
    assert not type_matches([1, 2, 3], "List[complex]")
    assert not type_matches([1 + 1j, 2.0, 3], "List[complex]")
    assert not type_matches(a, "List[str]")
    assert not type_matches(a, "List[int]")
    assert not type_matches(a, "List[float]")


def test_type_matches_unexpected():
    with pytest.raises(ValueError):
        assert type_matches("example", "weird")
        assert type_matches("example", "List[strange]")



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Validation facilities."""

import json
from pathlib import Path
from typing import Union

from .exceptions import ParselglossyError, collate_errors
from .utils import ComplexEncoder, JSONDict, path_resolver
from .validation_plumbing import (
    _rec_check_predicates,
    _rec_fix_defaults,
    _rec_merge_ours,
)
from .views import view_by_default, view_by_predicates, view_by_type


def validate_from_dicts(
    *, ir: JSONDict, template: JSONDict, fr_file: Union[str, Path] = None
) -> JSONDict:
    """Validate intermediate representation into final representation.

    Parameters
    ----------
    ir : JSONDict
        Intermediate representation of the input file.
    template : JSONDict
        A _validated_ template.
    fr_file : Union[str, Path]
         File to write final representation to (JSON format).
         None by default, which means file is not written out.

    Returns
    -------
    fr : JSONDict
        The validated input.

    Raises
    ------
    :exc:`ParselglossyError`
    """
    stencil = view_by_default(template)
    types = view_by_type(template)
    predicates = view_by_predicates(template)

    fr = merge_ours(theirs=stencil, ours=ir)
    fr = fix_defaults(fr, types=types)
    check_predicates(fr, predicates=predicates)

    if fr_file is not None:
        fr_file = path_resolver(fr_file)
        with fr_file.open("w") as out:
            json.dump(fr, out, cls=ComplexEncoder, indent=4)

    return fr


def merge_ours(*, theirs: JSONDict, ours: JSONDict) -> JSONDict:
    """Recursively merge two `dict`-s with "ours" strategy.

    Parameters
    ----------
    theirs : JSONDict
    ours : JSONDict

    Returns
    -------
    outgoing : JSONDict

    Raises
    ------
    :exc:`ParselglossyError`

    Notes
    -----
    The ``theirs`` dictionary is supposed to be the view by defaults of the
    validation specification, whereas ``ours`` is the dictionary from user
    input. The recursive merge action will generate a complete, but not
    validated, input dictionary by using default values where these are not
    overridden by user input, hence the naming "ours" for the merge strategy.

    This is porcelain over the recursive function :func:`_rec_merge_ours`.
    """
    outgoing, errors = _rec_merge_ours(theirs=theirs, ours=ours)

    if errors:
        msg = collate_errors(when="merging", errors=errors)
        raise ParselglossyError(msg)

    return outgoing


def fix_defaults(incoming: JSONDict, *, types: JSONDict) -> JSONDict:
    """Fix default values and perform type checking.

    Parameters
    ----------
    incoming: JSONDict
        The input ``dict``. This is supposed to be the one obtained by merging
        user and template ``dict``-s.
    types: JSONDict
        Types of all keywords in the input. Generated from :func:`view_by_types`.

    Returns
    -------
    outgoing: JSONDict
        A dictionary with all default values fixed.

    Raises
    ------
    :exc:`ParselglossyError`

    Notes
    -----
    Since we allow callables to appear as defaults, we need to run them
    to determine the actual default values.

    This operation must be done with some care, to avoid false negatives or
    ambiguous type checks. For example:

    * If the type is ``str`` and the default a callable, the type will
      match, but the default will make no sense.
    * If the type is numerical, *e.g.* ``int``, the type will not match.

    However, by design the callables **must** refer to some other field in the
    input tree, hence they **must** contain the reserved token "user". This
    allows us to disambiguate a callable as default from a value as default.

    The final strategy adopted is then:

    1. Perform type checking with :func:`type_matches`. If successful, we
    coerce the type.
    2. If types did not match, we further check whether the value is a string,
    containing the reserved token "value". This means the default value is
    actually a callable. We run the callable, which internally coerces the type
    of the result to the expected one.
    3. If even this check was unsuccessful, types really were unmatched. We
    report the error and move on.

    This is porcelain over recursive function :func:`_rec_fix_defaults`.
    """

    outgoing, errors = _rec_fix_defaults(incoming, types=types)

    if errors:
        msg = collate_errors(when="fixing defaults", errors=errors)
        raise ParselglossyError(msg)

    return outgoing


def check_predicates(incoming: JSONDict, *, predicates: JSONDict) -> None:
    """Run predicates on input tree with fixed defaults.

    Parameters
    ----------
    incoming : JSONDict
        The input ``dict``. This is supposed to be the result of :func:`fix_defaults`.
    predicates : JSONDict
        A view-by-predicates of the template ``dict``.

    Raises
    ------
    :exc:`ParselglossyError`

    Notes
    -----
    This is porcelain over recursive function :func:`_rec_check_predicates`.
    """

    errors = _rec_check_predicates(incoming, predicates=predicates)

    if errors:
        msg = collate_errors(when="checking predicates", errors=errors)
        raise ParselglossyError(msg)



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Top-level functions for parselglossy."""

import json
from pathlib import Path
from typing import List, Optional, Union

from pyparsing import __file__ as ppfile

from . import generation
from .documentation import documentation_generator
from .exceptions import ParselglossyError
from .grammars import lexer
from .utils import JSONDict, as_complex, copier, path_resolver
from .check_template import is_template_valid
from .validation import validate_from_dicts
from .yaml_utils import read_yaml_file


def generate(
    template: Union[str, Path],
    *,
    where: Union[str, Path] = Path.cwd() / "input_parser",
    grammar: Union[str, Path, List[Path]] = "standard",
    tokenize: Optional[str] = None,
    docfile: str = "input.rst",
    doc_header: str = "Input parameters",
) -> Path:
    """Generate parser for client.

    Parameters
    ----------
    template : Union[str, Path]
        Which validation template to use.
    where : Union[str, Path]
        Where to generate the parsing module. Default to ``input_parser`` folder
        under current working directory.
    grammar : Union[str, Path, List[Path]]
        The file containing the grammar to use to tokenize user input.
        Defaults to ``standard``, *i.e.* use one of the grammars packaged with
        the library, based on ``pyparsing``.
    tokenize : Optional[str]
        The commands to perform lexing of the input with a custom grammar. The
        result of these commands must be a variable named ``ir`` of type
        ``Dict[str, Any]``.
        Defaults to ``None``.
    docfile : str
        The name of the documentation file for the input.
        Defaults to ``input.rst``.
    doc_header : str
        Header for the documentation page.
        Defaults to ``Input parameters``.

    Returns
    -------
    Location of generated parser Python module as a ``Path`` object.

    Notes
    -----
    This function will generate a Python module for parsing inputs as defined
    by the template and grammar provided as parameters.

    The user can provide a grammar as (a list of) external file(s).
    There are a few constraints:

    * The function performing tokenization **must return** an object of
      ``Dict[str, Any]`` type: a, potentially recursive, dictionary of keys and
      values.
    * Commands to perform tokenization are passed *via* the ``tokenize`` input
      parameters. This is a string that will copied *verbatim* in the
      generated code, relevant ``import`` statements have to be included in
      this string!
    * The result of the commands in ``tokenize`` must be a variable named
      ``ir``.
    * It is the responsibility of the user to satisfy any dependencies of
      the parsing grammar they provided.

    Raises
    ------
    :exc:`ParselglossyError`
    """

    where_ = path_resolver(where, touch=False)

    # validate grammar options
    if isinstance(grammar, (Path, list)):
        if tokenize is None:
            raise ParselglossyError(
                "You need to set custom lexing commands when using a custom grammar."
            )
        else:
            lexer_str = tokenize

    # Create plumbing subfolder
    (where_ / "plumbing").mkdir(parents=True, exist_ok=True)
    # Generate __init__.py
    with (where_ / "__init__.py").open("w") as f:
        f.write(generation.INIT_PY)
    # Generate README.md
    with (where_ / "README.md").open("w") as f:
        f.write(generation.README)
    # Copy files
    # b. copy exceptions.py, types.py, utils.py, validation.py,
    # validation_plumbing.py, views.py file
    for fname in [
        "exceptions",
        "types",
        "utils",
        "validation",
        "validation_plumbing",
        "views",
    ]:
        copier(Path(__file__).parent.absolute() / f"{fname}.py", where_ / "plumbing")
    # c. Copy the grammar.
    # By default this is our standard grammar, but could be one file provided
    # by the user. The file is called "grammar.py"
    if isinstance(grammar, str):
        if grammar not in ["standard", "getkw"]:
            raise ParselglossyError(f"Grammar {grammar} not available.")
        for x in [
            ppfile,
            Path(__file__).parent.absolute() / "grammars/atoms.py",
            Path(__file__).parent.absolute() / "grammars/getkw.py",
        ]:
            copier(x, where_ / "plumbing")
        # extract `parse_string_to_dict` and write it into `getkw.py`
        fun = generation.get_parse_string_to_dict()
        with (where_ / "plumbing/getkw.py").open("a") as buf:
            buf.write(fun)
        lexer_str = f"""from . import getkw
    lexer = getkw.grammar(has_complex={grammar == 'standard'})
    ir = getkw.parse_string_to_dict(lexer, in_str)"""
    elif isinstance(grammar, Path):
        copier(grammar.resolve(), where_ / "plumbing/grammar.py")
    elif isinstance(grammar, list):
        for _ in grammar:
            copier(_.resolve(), where_ / "plumbing")
    else:
        raise ParselglossyError("Type not recognized for grammar")

    # d. generate lexer.py
    with (where_ / "plumbing/lexer.py").open("w") as f:
        f.write(generation.lexer_py(lexer_str))

    # Generate api.py
    template_ = path_resolver(template, touch=False)
    stencil = read_yaml_file(template_)

    # Check if the template is valid
    stencil = is_template_valid(stencil)

    with (where_ / "api.py").open("w") as f:
        f.write(generation.api_py(stencil))
    # Generate cli.py (uses argparse)
    with (where_ / "cli.py").open("w") as f:
        f.write(generation.CLI_PY)
    # Generate documentation
    (where_ / "docs").mkdir(parents=True, exist_ok=True)
    docs = documentation_generator(stencil, header=doc_header)
    outfile = path_resolver(where_ / f"docs/{docfile}")
    with outfile.open("w") as o:
        o.write(docs)

    return where_


def lex(
    infile: Union[str, Path],
    grammar: str = "standard",
    ir_file: Optional[Union[str, Path]] = None,
) -> JSONDict:
    """Run grammar of choice on input string.

    Parameters
    ----------
    infile : Union[str, Path]
        The string to be parsed.
    grammar : str
        Grammar to be used.
    ir_file : Union[str, Path]
        File to write intermediate representation to (JSON format).
        None by default, which means file is not written out.

    Returns
    -------
    The contents of the input string as a dictionary.
    """

    infile = path_resolver(infile)
    if ir_file is not None:
        ir_file = path_resolver(ir_file)

    with infile.open("r") as f:
        ir = lexer.lex_from_str(in_str=f.read(), grammar=grammar, ir_file=ir_file)

    return ir


def validate(
    infile: Union[str, Path],
    template: Union[str, Path],
    fr_file: Optional[Union[str, Path]] = None,
) -> JSONDict:
    """Validate intermediate representation into final representation.

    Parameters
    ----------
    infile : Union[str, Path]
        The file with the intermediate representation (JSON format).
    template : Union[str, Path]
        Which validation template to use.
    fr_file : Union[str, Path]
        File to write final representation to (JSON format).
        None by default, which means file is not written out.

    Returns
    -------
    The validated input as a dictionary.
    """

    infile = path_resolver(infile)
    with infile.open("r") as f:
        ir = json.load(f, object_hook=as_complex)

    template = path_resolver(template)
    stencil = read_yaml_file(template)

    if fr_file is not None:
        fr_file = path_resolver(fr_file)

    stencil = is_template_valid(stencil)

    return validate_from_dicts(ir=ir, template=stencil, fr_file=fr_file)


def parse(
    infile: Union[str, Path],
    template: Union[str, Path],
    outfile: Optional[Union[str, Path]] = None,
    grammar: str = "standard",
    dump_ir: bool = False,
) -> JSONDict:
    """Parse input file.

    Parameters
    ----------
    infile : Union[str, Path]
        The input file to be parsed.
    template : Union[str, Path]
        Which validation template to use.
    outfile : Optional[Union[str, Path]]
        The output file.
        Defaults to ``None``, which means writing to ``<infile>_fr.json``.
    grammar : str
        Which grammar to use. Defaults to ``standard``.
    dump_ir : bool
        Whether to write out the intermediate representation to file (JSON format).
        False by default. If true the filename if ``<infile>_ir.json``

    Returns
    -------
    The validated input as a dictionary.
    """

    infile = path_resolver(infile)

    ir_file = None  # type: Optional[Path]
    if dump_ir:
        ir_file = path_resolver(Path(infile).stem + "_ir.json")

    ir = lex(infile=infile, ir_file=ir_file, grammar=grammar)

    template = path_resolver(template)
    stencil = read_yaml_file(template)

    if outfile is not None:
        outfile = path_resolver(outfile)

    stencil = is_template_valid(stencil)

    return validate_from_dicts(ir=ir, template=stencil, fr_file=outfile)


def document(
    template: Union[str, Path],
    outfile: Optional[Union[str, Path]] = None,
    header: str = "Input parameters",
) -> str:
    """Generate documentation in reStructuredText format from validation template.

    Parameters
    ----------
    template : Union[str, Path]
        Which validation template to use.
    output : Union[str, Path]
        Where to save the generated documentation.
        Defaults to ``None``.
    header : str
        Header for the documentation page.
        Defaults to ``Input parameters``.

    Returns
    -------
    The documentation page as a string.
    """

    template = path_resolver(template)
    stencil = read_yaml_file(template)

    stencil = is_template_valid(stencil)

    docs = documentation_generator(stencil, header=header)

    if outfile is not None:
        outfile = path_resolver(outfile)
        with outfile.open("w") as o:
            o.write(docs)

    return docs



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Tests for `parselglossy` package."""

import json
from pathlib import Path
from shutil import rmtree

import pytest

from parselglossy import api
from parselglossy.utils import as_complex


@pytest.mark.parametrize(
    "args,out,reference",
    [
        (("tests/api/standard.inp",), None, "tests/ref/standard_ir.json"),
        (
            ("tests/api/scf.inp", "standard", "scf_ir.json"),
            "scf_ir.json",
            "tests/ref/scf_ir.json",
        ),
        (
            ("tests/api/getkw.inp", "getkw", "lex_ir.json"),
            "lex_ir.json",
            "tests/ref/getkw_ir.json",
        ),
    ],
    ids=["lex-default", "lex-standard-w-outfile", "lex-getkw-w-outfile"],
)
def test_api_lex(args, out, reference):
    ir = api.lex(*args)

    # Check intermediate representation matches with reference
    ref_json = Path(reference).resolve()
    if out is None:
        with ref_json.open("r") as f:
            assert ir == json.loads(f.read(), object_hook=as_complex)
    else:
        dumped = Path(out).resolve()
        with ref_json.open("r") as ref, dumped.open("r") as o:
            assert json.loads(o.read(), object_hook=as_complex) == json.loads(
                ref.read(), object_hook=as_complex
            )
        # Clean up JSON file
        dumped.unlink()


@pytest.mark.parametrize(
    "args,out,reference",
    [
        (
            ("tests/ref/scf_ir.json", "tests/validation/overall/template.yml"),
            None,
            "tests/ref/scf_fr.json",
        ),
        (
            (
                "tests/ref/scf_ir.json",
                "tests/validation/overall/template.yml",
                "scf_fr.json",
            ),
            "scf_fr.json",
            "tests/ref/scf_fr.json",
        ),
    ],
    ids=["validate-default", "validate-w-outfile"],
)
def test_api_validate(args, out, reference):
    fr = api.validate(*args)

    # Check final representation matches with reference
    ref_json = Path(reference).resolve()
    if out is None:
        with ref_json.open("r") as f:
            assert fr == json.loads(f.read(), object_hook=as_complex)
    else:
        dumped = Path(out).resolve()
        with ref_json.open("r") as ref, dumped.open("r") as o:
            assert json.loads(o.read(), object_hook=as_complex) == json.loads(
                ref.read(), object_hook=as_complex
            )
        # Clean up JSON file
        dumped.unlink()


@pytest.mark.parametrize(
    "args,out,reference",
    [
        (
            ("tests/api/scf.inp", "tests/validation/overall/template.yml"),
            None,
            "tests/ref/scf_fr.json",
        ),
        (
            (
                "tests/api/scf.inp",
                "tests/validation/overall/template.yml",
                "scf_fr.json",
            ),
            "scf_fr.json",
            "tests/ref/scf_fr.json",
        ),
        (
            (
                "tests/api/scf.inp",
                "tests/validation/overall/template.yml",
                None,
                "standard",
                True,
            ),
            None,
            "tests/ref/scf_fr.json",
        ),
    ],
    ids=["parse-default", "parse-w-outfile", "parse-w-dumpir"],
)
def test_api_parse(args, out, reference):
    fr = api.parse(*args)

    # Check final representation matches with reference
    ref_json = Path(reference).resolve()
    if out is None:
        with ref_json.open("r") as f:
            assert fr == json.loads(f.read(), object_hook=as_complex)
    else:
        dumped = Path(out).resolve()
        with ref_json.open("r") as ref, dumped.open("r") as o:
            assert json.loads(o.read(), object_hook=as_complex) == json.loads(
                ref.read(), object_hook=as_complex
            )
        # Clean up JSON file
        dumped.unlink()


@pytest.mark.parametrize(
    "args,out,reference",
    [
        (("tests/api/docs_template.yml",), None, "tests/ref/input.rst"),
        (
            ("tests/api/docs_template.yml", "input.rst"),
            "input.rst",
            "tests/ref/input.rst",
        ),
        (
            (
                "tests/api/docs_template.yml",
                "input.rst",
                "Dwigt Rortugal's guide to input parameters",
            ),
            "input.rst",
            "tests/ref/dwigt.rst",
        ),
    ],
    ids=["document-default", "document-w-outfile", "document-w-header"],
)
def test_api_document(args, out, reference):
    docs = api.document(*args)

    # Check final representation matches with reference
    ref_rst = Path(reference).resolve()
    if out is None:
        with ref_rst.open("r") as ref:
            assert docs == ref.read()
    else:
        dumped = Path(out).resolve()
        with ref_rst.open("r") as ref, dumped.open("r") as o:
            assert o.read() == ref.read()

        # Clean up .rst file
        dumped.unlink()


@pytest.mark.parametrize(
    "args,inp,references",
    [
        (
            {
                "template": "tests/validation/overall/template.yml",
                "docfile": "foo.rst",
                "grammar": "standard",
            },
            "tests/api/scf.inp",
            ("tests/ref/scf_fr.json", "tests/ref/generated_input.rst"),
        ),
        (
            {
                "template": "tests/validation/overall/template.yml",
                "docfile": "babar.rst",
                "grammar": [
                    Path(__file__).parents[1].absolute()
                    / f"parselglossy/grammars/{x}.py"
                    for x in ["atoms", "getkw"]
                ],
                "tokenize": "from . import getkw; lexer = getkw.grammar(has_complex=True); ir = lexer.parseString(in_str).asDict()",
            },
            "tests/api/scf.inp",
            ("tests/ref/scf_fr.json", "tests/ref/generated_input.rst"),
        ),
    ],
    ids=["generate-default", "generated-custom-grammar"],
)
def test_api_generate(args, inp, references):
    # generate parser
    parser_dir = api.generate(**args)

    # use generated parser to parse inp
    from input_parser import api as generated_api

    fr = generated_api.parse(inp)

    # Check final representation matches with reference
    ref_json = Path(references[0]).resolve()
    with ref_json.open("r") as ref:
        assert fr == json.loads(ref.read(), object_hook=as_complex)

    # Check generated documentation agrees with reference
    ref_rst = Path(references[1]).resolve()
    dumped = parser_dir / f"docs/{args['docfile']}"
    with ref_rst.open("r") as ref, dumped.open("r") as o:
        assert o.read() == ref.read()

    # clean up generated parser
    rmtree(parser_dir)



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

from pathlib import Path

import pytest

from parselglossy.documentation import documentation_generator


@pytest.fixture
def template():
    return {
        "keywords": [
            {"name": "title", "type": "str", "docstring": "Title of the calculation."}
        ],
        "sections": [
            {
                "name": "foo",
                "docstring": "brilliant",
                "keywords": [
                    {
                        "name": "tragic",
                        "type": "str",
                        "docstring": "Title of the calculation.",
                        "default": "Wham! Bang! Pow! Let's Rock Out!",
                    }
                ],
                "sections": [
                    {
                        "name": "Bar",
                        "docstring": "A ba-bar section",
                        "keywords": [
                            {
                                "name": "amazing",
                                "type": "str",
                                "docstring": "Title of the calculation.",
                                "default": "Bobson Dugnutt",
                            }
                        ],
                    }
                ],
            }
        ],
    }


def test_documentation(template):
    doc_ref = Path(__file__).parent / Path("ref/simple.rst")
    with doc_ref.open("r") as ref:
        stuff = ref.read().rstrip("\n")

    assert documentation_generator(template) == stuff



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Tests for `parselglossy` package.

Tests that the template is not malformed.
"""

from contextlib import ExitStack as does_not_raise
from copy import deepcopy

import pytest

from parselglossy.exceptions import ParselglossyError
from parselglossy.check_template import is_template_valid

# Well formed keyword and section
keyword = {"name": "title", "type": "str", "docstring": "Title of the calculation."}
section = {"name": "scf", "docstring": "SCF input parameters.", "keywords": [keyword]}

valid = {
    "keywords": [
        {"name": "title", "type": "str", "docstring": "Title of the calculation."}
    ],
    "sections": [
        {
            "name": "foo",
            "docstring": "brilliant",
            "keywords": [
                {
                    "name": "tragic",
                    "type": "str",
                    "docstring": "Title of the calculation.",
                }
            ],
        }
    ],
}

# Malformed keywords
keyword_with_section = deepcopy(keyword)
keyword_with_section["sections"] = [deepcopy(section)]

keyword_with_invalid_type = deepcopy(keyword)
keyword_with_invalid_type["type"] = "fooffa"

keyword_without_type = deepcopy(keyword)
keyword_without_type.pop("type")

keyword_without_docstring = deepcopy(keyword)
keyword_without_docstring.pop("docstring")

keyword_with_nothing = deepcopy(keyword_without_type)
keyword_with_nothing.pop("docstring")

# Malformed sections
section_without_docstring = deepcopy(section)
section_without_docstring.pop("docstring")

nested_sections_without_docstring = {
    "sections": [
        {
            "name": "foo",
            "keywords": [
                {
                    "name": "title",
                    "type": "str",
                    "docstring": "Title of the calculation.",
                }
            ],
            "sections": [{"name": "bar"}],
        }
    ]
}

nested_sections_with_malformed_keyword = {
    "sections": [
        {
            "name": "foo",
            "keywords": [keyword_with_section, keyword_with_nothing],
            "sections": [{"name": "bar"}],
        }
    ]
}

cycles = {
    "sections": [
        {
            "docstring": "Section docs",
            "keywords": [
                {
                    "default": "user['some_section']['another_number']",
                    "docstring": "Some number which defaults to the "
                    "value of another_number.\n",
                    "name": "some_number",
                    "type": "int",
                },
                {
                    "default": "user['some_section']['some_number']",
                    "docstring": "Another number which defaults to "
                    "the value of some_number.\n",
                    "name": "another_number",
                    "type": "int",
                },
            ],
            "name": "some_section",
        },
        {
            "docstring": "Section docs",
            "keywords": [
                {
                    "default": "user['another_section']['another_number']",
                    "docstring": "Some number which defaults to the "
                    "value of another_number.\n",
                    "name": "some_number",
                    "type": "int",
                },
                {
                    "default": "user['another_section']['some_number']",
                    "docstring": "Another number which defaults to "
                    "the value of some_number.\n",
                    "name": "another_number",
                    "type": "int",
                },
            ],
            "name": "another_section",
        },
    ],
}


error_preamble = r"Error(?:s)? occurred when checking the template:\n"
check_template_data = [
    (valid, does_not_raise()),
    (
        {"keywords": [keyword_with_section]},
        pytest.raises(
            ParselglossyError,
            match=(
                error_preamble
                + r"- At user\['title'\]:\s+Sections cannot be nested under keywords."
            ),
        ),
    ),
    (
        {"keywords": [keyword_with_invalid_type]},
        pytest.raises(
            ParselglossyError,
            match=(
                error_preamble
                + r"- At user\['title'\]:\s+Keywords must have a valid type."
            ),
        ),
    ),
    (
        {"keywords": [keyword_without_type]},
        pytest.raises(
            ParselglossyError,
            match=(
                error_preamble
                + r"- At user\['title'\]:\s+Keywords must have a valid type."
            ),
        ),
    ),
    (
        {"keywords": [keyword_without_docstring]},
        pytest.raises(
            ParselglossyError,
            match=(
                error_preamble
                + r"- At user\['title'\]:\s+Keywords must have a non-empty docstring."
            ),
        ),
    ),
    (
        {"keywords": [keyword_with_nothing]},
        pytest.raises(
            ParselglossyError,
            match=(
                error_preamble
                + r"- At user\['title'\]:\s+Keywords must have a valid type.\n"
                r"- At user\['title'\]:\s+Keywords must have a non-empty docstring."
            ),
        ),
    ),
    (
        {"sections": [section_without_docstring]},
        pytest.raises(
            ParselglossyError,
            match=(
                error_preamble
                + r"- At user\['scf'\]:\s+Sections must have a non-empty docstring."
            ),
        ),
    ),
    (
        nested_sections_without_docstring,
        pytest.raises(
            ParselglossyError,
            match=(
                error_preamble
                + r"- At user\['foo'\]:\s+Sections must have a non-empty docstring.\n"
                r"- At user\['foo'\]\['bar'\]:\s+Sections must have a non-empty docstring."
            ),
        ),
    ),
    (
        nested_sections_with_malformed_keyword,
        pytest.raises(
            ParselglossyError,
            match=(
                error_preamble
                + r"- At user\['foo'\]:\s+Sections must have a non-empty docstring.\n"
                r"- At user\['foo'\]\['title'\]:\s+Sections cannot be nested under keywords.\n"
                r"- At user\['foo'\]\['title'\]:\s+Keywords must have a valid type.\n"
                r"- At user\['foo'\]\['title'\]:\s+Keywords must have a non-empty docstring.\n"
                r"- At user\['foo'\]\['bar'\]:\s+Sections must have a non-empty docstring."
            ),
        ),
    ),
    (
        cycles,
        pytest.raises(
            ParselglossyError,
            match=(
                error_preamble
                + r"- At user\['(some|another)_section'\]\['(some|another)_number'\]:\s+Keyword depends cyclically on keyword user\['(some|another)_section'\]\['(some|another)_number'\]\n"
                r"- At user\['(some|another)_section'\]\['(some|another)_number'\]:\s+Keyword depends cyclically on keyword user\['(some|another)_section'\]\['(some|another)_number'\]"
            ),
        ),
    ),
]


@pytest.mark.parametrize(
    "template,raises",
    check_template_data,
    ids=[
        "valid",
        "section_under_keyword",
        "keyword_no_type",
        "keyword_invalid_type",
        "keyword_no_docstring",
        "keyword_nothing",
        "section_no_docstring",
        "nested_no_docstring",
        "nested_bad_keyword",
        "cycles",
    ],
)
def test_check_template(template, raises):
    with raises:
        template = is_template_valid(template)



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Documentation generation."""

from typing import List  # noqa: F401

from .utils import JSONDict


def documentation_generator(
    template: JSONDict, *, header: str = "Input parameters"
) -> str:
    """Generates documentation from a valid template.

    Parameters
    ----------
    template : JSONDict
        The template to generate documentation from.
        We assume that the template is valid.

    Returns
    -------
    documentation : str
    """

    comment = (
        ".. raw:: html\n\n"  # noqa: F541
        "    <style> .red {color:#aa0060; font-weight:bold; font-size:18px} </style>\n\n"  # noqa: E501
        ".. role:: red\n\n"
        f".. This documentation was autogenerated using parselglossy."
        " Editing by hand is not recommended.\n"
    )

    header = (
        f"{comment:s}\n{'=' * len(header):s}\n{header:s}\n{'=' * len(header):s}\n\n"
        "- Keywords without a default value are **required**.\n"
        "- Default values are either explicit or computed from the value of other keywords in the input.\n"  # noqa: E501
        "- Sections where all keywords have a default value can be omitted.\n"
        "- Predicates, if present, are the functions run to validate user input.\n"
    )

    docs = _rec_documentation_generator(template=template)

    documentation = header + docs

    return documentation


def _document_keyword(keyword: JSONDict) -> str:
    docstring = keyword["docstring"].replace("\n", " ")
    doc = f"""
 :{keyword['name']:s}: {docstring:s}

  **Type** ``{keyword['type']:s}``
"""

    if "default" in keyword.keys():
        doc += f"""
  **Default** ``{keyword['default']}``
"""

    if "predicates" in keyword.keys():
        preds = "\n    ".join((f"- ``{x}``" for x in keyword["predicates"]))
        doc += f"""
  **Predicates**
    {preds}
"""

    return doc


def _rec_documentation_generator(template, *, level: int = 0) -> str:
    """Generates documentation from a valid template.

    Parameters
    ----------
    template : JSONDict
    level : int

    Returns
    -------
    docs : str
    """

    docs = []  # type: List[str]

    keywords = template["keywords"] if "keywords" in template.keys() else []
    if keywords:
        docs.append(_indent("\n:red:`Keywords`", level))

    for k in keywords:
        doc = _document_keyword(k)
        docs.extend(_indent(doc, level))

    sections = template["sections"] if "sections" in template.keys() else []
    if sections:
        docs.append(_indent("\n:red:`Sections`", level))
        for s in sections:
            docstring = s["docstring"].replace("\n", " ")
            doc = f"\n :{s['name']:s}: {docstring:s}\n"
            doc += _rec_documentation_generator(s, level=level + 1)
            docs.extend(_indent(doc, level))

    return "".join(docs)


def _indent(in_str: str, level: int = 0) -> str:
    return in_str.replace("\n", "\n" + ("  " * level))



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Tests for `parselglossy` package.

Tests defaulting:
  - Required keywords that are absent report errors.
  - Invalid callables report errors.

To test, we create a fixture with the types and a valid reference dictionary.
We then proceed to create copies of the reference and modify some of its
values, to see whether we fix back defaults correctly.
"""

import re
from contextlib import ExitStack as does_not_raise
from copy import deepcopy

import pytest

from parselglossy.exceptions import ParselglossyError
from parselglossy.validation import fix_defaults


def valid():
    return {
        "scf": {
            "another_number": 10,
            "functional": "B3LYP",
            "max_num_iterations": 20,
            "some_acceleration": False,
            "list_complex_number": [complex("0.0+0.0j"), complex("1.0+0.5j")],
            "thresholds": {
                "some_complex_number": complex("0.0+0.0j"),
                "energy": 0.001,
                "some_integral_screening": 0.0001,
            },
        },
        "list_of_strings": ["foo", "bar"],
        "title": "My fantastic calculation",
    }


@pytest.fixture
def types():
    return {
        "scf": {
            "another_number": "int",
            "functional": "str",
            "max_num_iterations": "int",
            "some_acceleration": "bool",
            "list_complex_number": "List[complex]",
            "thresholds": {
                "some_complex_number": "complex",
                "energy": "float",
                "some_integral_screening": "float",
            },
        },
        "list_of_strings": "List[str]",
        "title": "str",
    }


raw = {
    "scf": {
        "another_number": 10,
        "functional": "B3LYP",
        "max_num_iterations": 20,
        "some_acceleration": False,
        "list_complex_number": ["0.0+0.0j", "1.0+0.5j"],
        "thresholds": {
            "some_complex_number": "0.0+0.0j",
            "energy": 0.001,
            "some_integral_screening": 0.0001,
        },
    },
    "list_of_strings": ["foo", "bar"],
    "title": "My fantastic calculation",
}


def valid_with_action():
    d = deepcopy(raw)
    d["scf"]["another_number"] = "user['scf']['max_num_iterations'] / 10"
    r = valid()
    r["scf"]["another_number"] = 2
    return d, r


def one_invalid_action():
    d = deepcopy(raw)
    d["scf"]["another_number"] = "user['scf']['min_num_iterations'] / 2"
    return d


def type_error_action():
    d = deepcopy(raw)
    d["scf"]["another_number"] = "user['title'] / 2"
    return d


def two_invalid_actions():
    d = deepcopy(raw)
    d["scf"]["another_number"] = "user['scf']['min_num_iterations'] / 2"
    d["scf"]["thresholds"]["energy"] = "10 * False"
    return d


error_preamble = r"Error(?:s)? occurred when fixing defaults:\n"

msg1 = r"- At user\['scf'\]\['another_number'\]:\s+KeyError 'min_num_iterations' in closure 'user\['scf'\]\['min_num_iterations'\] / 2'\."

msg2 = r"- At user\['scf'\]\['thresholds'\]\['energy'\]:\s+Actual \(str\) and declared \(float\) types do not match\."

# We need to check against two possible orderings of the error message.
# In Python 3.5 orderings of dict is not guaranteed!
invalid_messages = [
    error_preamble + msg1 + r"\n" + msg2,
    error_preamble + msg2 + r"\n" + msg1,
]

msg3 = r"- At user\['scf'\]\['another_number'\]:\s+TypeError unsupported operand type\(s\) for /: 'str' and 'int' in closure 'user\['title'\] / 2'\."

testdata = [
    (raw, valid(), does_not_raise(), [""]),
    (*valid_with_action(), does_not_raise(), [""]),
    (
        one_invalid_action(),
        valid(),
        pytest.raises(ParselglossyError),
        [error_preamble + msg1],
    ),
    (
        type_error_action(),
        valid(),
        pytest.raises(ParselglossyError),
        [error_preamble + msg3],
    ),
    (
        two_invalid_actions(),
        valid(),
        pytest.raises(ParselglossyError),
        invalid_messages,
    ),
]


@pytest.mark.parametrize(
    "user,ref,raises,error_message",
    testdata,
    ids=[
        "noactions",
        "valid_action",
        "one_invalid_action",
        "type_error_action",
        "two_invalid_actions",
    ],
)
def test_fix_defaults(types, user, ref, raises, error_message):
    with raises as e:
        outgoing = fix_defaults(user, types=types)
        assert outgoing == ref
        # Check error message is correct
        assert re.match("|".join(error_message), str(e)) is not None



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Facilities for checking input templates"""

import re
from copy import deepcopy
from typing import Tuple, List, Any

import networkx as nx

from .exceptions import Error, ParselglossyError, collate_errors
from .utils import JSONDict, location_in_dict
from .types import allowed_types
from .views import view_by_default, view_by_default_keywords


def is_template_valid(template: JSONDict) -> JSONDict:
    """Checks a template ``dict`` is well-formed.

    A template ``dict`` is well-formed if:

    * All keywords have:

      - An allowed type.
      - A non-empty docstring.
      - A default callable that is valid Python, if present.
      - Predicates that are valid Python, if present.

      Note that the latter two criteria can only be checked later on.

    * No sections are nested under keywords.

    * All sections have a non-empty docstring.

    * Reorders keywords according to their dependencies.

    Parameters
    ----------
    template : JSONDict

    Returns
    -------
    ordered : JSONDict

    Raises
    ------
    :exc:`ParselglossyError`

    Notes
    -----
    This is porcelain over the recursive :func:`_rec_is_template_valid`.
    """

    errors = _rec_is_template_valid(template)
    errors.extend(_check_cyclic_defaults(template))

    if errors:
        msg = collate_errors(when="checking the template", errors=errors)
        raise ParselglossyError(msg)

    return _reorder_template(template)


def _rec_is_template_valid(template: JSONDict, *, address: Tuple = ()) -> List[Error]:
    """Checks a template ``dict`` is well-formed.

    A template ``dict`` is well-formed if:

    * All keywords have:

      - An allowed type.
      - A non-empty docstring.
      - A default callable that is valid Python, if present.
      - Predicates that are valid Python, if present.

      Note that the latter two criteria can only be checked later on.

    * No sections are nested under keywords.

    * All sections have a non-empty docstring.

    Parameters
    ----------
    template : JSONDict
    address : Tuple[str]

    Returns
    -------
    errors : List[Error]
    """

    errors = []

    keywords = template["keywords"] if "keywords" in template.keys() else []
    for k in keywords:
        errs = _check_keyword(k, address=address)
        errors.extend(errs)

    sections = template["sections"] if "sections" in template.keys() else []
    for s in sections:
        if _undocumented(s):
            errors.append(
                Error(
                    (address + (s["name"],)),
                    "Sections must have a non-empty docstring.",
                )
            )
        errs = _rec_is_template_valid(s, address=(address + (s["name"],)))
        errors.extend(errs)

    return errors


def _check_cyclic_defaults(template: JSONDict) -> List[Error]:
    """Check for cyclic dependencies between defaulting actions."""
    stencil = view_by_default(template)
    dependencies = _sections_default_dependencies(stencil)
    # for nx we need to translate from tuples of lists to tuples of tuples
    dependencies_hashable = [
        (tuple(_from), tuple(_to)) for (_from, _to) in dependencies
    ]
    G = nx.DiGraph(dependencies_hashable)
    cycles = nx.simple_cycles(G)
    errors = []
    for c in cycles:
        errors.append(
            Error(
                c[0],
                f"Keyword depends cyclically on keyword {location_in_dict(address=c[1])}",  # noqa: E501
            )
        )
    return errors


def _sections_default_dependencies(
    section: JSONDict, parent_sections: List[str] = []
) -> List[Any]:
    """Collects a list of of tuples (keyword, default dependency).

    Both the keyword and the default dependency are lists.

    One such a tuple could look like this:
    (['main section', 'subsection', 'keyword'], ['main section', 'another keyword'])
    """
    dependencies = []
    for k, v in section.items():
        if not isinstance(v, dict):
            dependencies += _keywords_default_dependencies(k, v, parent_sections)
        else:
            dependencies += _sections_default_dependencies(v, parent_sections + [k])
    return dependencies


def _keywords_default_dependencies(
    keyword: str, default: str, parent_sections: List[str]
) -> List[Any]:
    """Collects a list of of tuples (keyword, default dependency) within one section.

    Both the keyword and the default dependency are lists.

    One such a tuple could look like this:
    (['main section', 'subsection', 'keyword'], ['main section', 'another keyword'])
    """
    dependencies = []
    if isinstance(default, str) and "user" in default:
        _to = re.findall(r'\[[\'"](.*?)[\'"]\]', default)
        if len(_to) > 0:
            _from = parent_sections + [keyword]
            dependencies.append((_from, _to))
    return dependencies


def _reorder_template(template: JSONDict) -> JSONDict:
    """Reorder template according to direct graph of keyword dependencies

    Parameters
    ----------
    template : JSONDict

    Returns
    -------
    ordered : JSONDict

    Notes
    -----
    This function reorders the template in place.

    Warnings
    --------
    We are assuming that there are no dependency cycles in the template. Call
    this function **after** :func:`_check_cyclic_defaults`.
    """

    ordered = deepcopy(template)
    _rec_reoder_template(ordered)
    return ordered


def _rec_reoder_template(template: JSONDict) -> None:
    """Recursive backend for :func:`_reorder_template`.

    Parameters
    ----------
    template : JSONDict
    """
    keywords = template["keywords"] if "keywords" in template.keys() else []
    kw_stencil = view_by_default_keywords(keywords)
    deps = _sections_default_dependencies(kw_stencil)
    deps_hashable = [(tuple(_from), tuple(_to)) for (_from, _to) in deps]
    nodes = reversed([x[-1] for x in nx.DiGraph(deps_hashable)])

    shuffle = []
    for node in nodes:
        for x in keywords:
            if x["name"] == node:
                shuffle.append(deepcopy(x))
                keywords.remove(x)
    keywords.extend(shuffle)

    sections = template["sections"] if "sections" in template.keys() else []
    for s in sections:
        _rec_reoder_template(s)


def _check_keyword(keyword: JSONDict, *, address: Tuple = ()) -> List[Error]:
    """Checks that a template keyword is well-formed.

    In this function, a template keyword is well-formed if it has:

    * An allowed type.
    * A non-empty docstring.

    The two additional criteria:

    * A default callable that is valid Python, if present.
    * Predicates that are valid Python, if present.

    can only be checked meaningfully later, as they might depend on
    keyword(s)/section(s) of the input that are only known after merging.

    Parameters
    ----------
    template : JSONDict
    address : Tuple

    Returns
    -------
    errors : List[Error]
    """

    errors = []

    k = keyword["name"]
    if "sections" in keyword.keys():
        errors.append(
            Error((address + (k,)), "Sections cannot be nested under keywords.")
        )

    if _untyped(keyword):
        errors.append(Error((address + (k,)), "Keywords must have a valid type."))

    if _undocumented(keyword):
        errors.append(
            Error((address + (k,)), "Keywords must have a non-empty docstring.")
        )

    return errors


def _undocumented(x: JSONDict) -> bool:
    return (
        True if "docstring" not in x.keys() or x["docstring"].strip() == "" else False
    )


def _untyped(x: JSONDict) -> bool:
    return True if "type" not in x.keys() or x["type"] not in allowed_types else False



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Tools to extract views of dictionaries."""

from typing import Any, Callable, List, Optional, Type

from .exceptions import ParselglossyError
from .utils import JSONDict


def view_by_type(d: JSONDict) -> JSONDict:
    """Partial application of :func:`view_by` for types.

    Parameters
    ----------
    d: JSONDict

    Returns
    -------
    outgoing: JSONDict
       A dictionary with a view by types.
    """
    return view_by("type", d, missing=ParselglossyError)


def view_by_default(d: JSONDict) -> JSONDict:
    """Partial application of :func:`view_by` for defaults.

    Parameters
    ----------
    d: JSONDict

    Returns
    -------
    outgoing: JSONDict
       A dictionary with a view by defaults.
    """
    return view_by("default", d)


def view_by_default_keywords(keywords: List[JSONDict]) -> JSONDict:
    """View by defaults only for lists of keywords.

    Parameters
    ----------
    keywords: List[JSONDict]

    Returns
    -------
    outgoing: JSONDict
       A dictionary with a view by defaults for the keywords.
    """
    return {v["name"]: v["default"] if "default" in v else None for v in keywords}


def view_by_docstring(d: JSONDict) -> JSONDict:
    """Partial application of :func:`view_by` for docstrings.

    Parameters
    ----------
    d: JSONDict

    Returns
    -------
    outgoing: JSONDict
       A dictionary with a view by docstrings.
    """

    def docstring_not_empty(x: Any, y: str) -> bool:
        """Check that a docstring is not empty."""
        return x[y].strip() != ""

    def docstring_rstrip(x: str) -> str:
        """Apply rstrip to a docstring"""
        return x.rstrip()

    return view_by(
        "docstring",
        d,
        predicate=docstring_not_empty,
        missing=ParselglossyError,
        transformer=docstring_rstrip,
    )


def view_by_predicates(d: JSONDict) -> JSONDict:
    """Partial application of :func:`view_by` for predicates.

    Parameters
    ----------
    d: JSONDict

    Returns
    -------
    outgoing: JSONDict
       A dictionary with a view by predicates.
    """
    return view_by("predicates", d)


def view_by(
    what: str,
    d: JSONDict,
    *,
    predicate: Callable[[Any, str], bool] = lambda x, y: True,
    missing: Optional[Type[Exception]] = None,
    transformer: Callable[[Any], Any] = lambda x: x,
) -> JSONDict:
    """Recursive decimation of a template into an input.

    Parameters
    ----------
    what : str
        What view to extract from the dictionary. Any of ``type``, ``default``,
        ``docstring`` or ``predicates`` is allowed.
    d: JSONDict
    predicate: Callable
       A predicate accepting two arguments.
    missing: Optional[Exception]
    transformer: Callable

    Returns
    -------
    outgoing: JSONDict
       A dictionary with the desired view.

    Raises
    ------
    exc:`ValueError` if `what` is not among the allowed views.
    """

    allowed = ["type", "default", "docstring", "predicates"]
    if what not in allowed:
        raise ValueError(
            f"Requested view {what:s} not among possible views ({allowed})"
        )

    # Determine whether we have only sections and only keywords
    has_keywords = True if "keywords" in d else False
    has_sections = True if "sections" in d else False

    view = {}  # type: JSONDict
    if has_keywords:
        view = {
            v["name"]: transformer(v[what])
            if all([what in v, predicate(v, what)])
            else missing
            for v in d["keywords"]
        }

    if has_sections:
        for section in d["sections"]:
            view[section["name"]] = view_by(
                what,
                section,
                predicate=predicate,
                missing=missing,
                transformer=transformer,
            )

    return view



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

import json
from contextlib import ExitStack as does_not_raise
from pathlib import Path
from typing import List

import pytest

from parselglossy.exceptions import ParselglossyError
from parselglossy.utils import as_complex
from parselglossy.check_template import is_template_valid
from parselglossy.validation import validate_from_dicts
from read_in import read_in


def valid():
    return {
        "title": "this is an example",
        "scf": {
            "functional": "B3LYP",
            "max_num_iterations": 20,
            "some_acceleration": True,
            "thresholds": {"some_integral_screening": 0.0002, "energy": 1e-05},
            "another_number": 10,
            "some_complex_number": complex(0.0, 0.0),
        },
    }


template_errors_data = [
    (
        "template_errors",
        "input.yml",
        "template_no_documentation.yml",
        ParselglossyError,
        None,
        [
            r"- At user\['some_section'\]:\s+Sections must have a non-empty docstring\.\n"
            r"- At user\['some_section'\]\['a_short_string'\]:\s+Keywords must have a non-empty docstring\."
        ],
    ),
    (
        "template_errors",
        "input.yml",
        "template_empty_documentation.yml",
        ParselglossyError,
        None,
        [
            r"- At user\['some_section'\]:\s+Sections must have a non-empty docstring\.\n"
            r"- At user\['some_section'\]\['a_short_string'\]:\s+Keywords must have a non-empty docstring\."
        ],
    ),
    (
        "template_errors",
        "input.yml",
        "template_invalid_predicate.yml",
        ParselglossyError,
        None,
        [
            r"- At user\['some_section'\]\['a_short_string'\]:\s+NameError name 'undefined' is not defined in closure '0 < len\(value\) <= undefined'\."
        ],
    ),
    (
        "template_errors",
        "input.yml",
        "template_w_cycles.yml",
        ParselglossyError,
        None,
        [
            r"Error(?:s)? occurred when checking the template:\s+- At user\['(some|another)_section'\]\['(some|another)_number'\]:\s+Keyword depends cyclically on keyword user\['(some|another)_section'\]\['(some|another)_number'\]"
        ],
    ),
]

input_errors_data = [
    (
        "input_errors",
        "unexpected_keyword.yml",
        "template.yml",
        ParselglossyError,
        None,
        [r"Error(?:s)? occurred when merging:\n- Found unexpected keyword: 'strange'"],
    ),
    (
        "input_errors",
        "unexpected_section.yml",
        "template.yml",
        ParselglossyError,
        None,
        [r"Error(?:s)? occurred when merging:\n- Found unexpected section: 'weird'"],
    ),
    (
        "input_errors",
        "input_missing_keyword.yml",
        "template.yml",
        ParselglossyError,
        None,
        [
            r"Error(?:s)? occurred when merging:\n- At user\['some_section'\]\['a_short_string'\]:\s+Keyword 'a_short_string' is required but has no value\."
        ],
    ),
    (
        "input_errors",
        "input_type_error_bool.yml",
        "template.yml",
        ParselglossyError,
        None,
        [
            r"Error(?:s)? occurred when fixing defaults:\n- At user\['some_section'\]\['some_feature'\]:\s+Actual \(int\) and declared \(bool\) types do not match\."
        ],
    ),
    (
        "input_errors",
        "input_type_error_float.yml",
        "template.yml",
        ParselglossyError,
        None,
        [
            r"Error(?:s)? occurred when fixing defaults:\n- At user\['some_float'\]:\s+Actual \(int\) and declared \(float\) types do not match\."
        ],
    ),
    (
        "input_errors",
        "input_type_error_int.yml",
        "template.yml",
        ParselglossyError,
        None,
        [
            r"Error(?:s)? occurred when fixing defaults:\n- At user\['some_section'\]\['some_number'\]:\s+Actual \(float\) and declared \(int\) types do not match\."
        ],
    ),
    (
        "input_errors",
        "input_type_error_list.yml",
        "template.yml",
        ParselglossyError,
        None,
        [
            r"Error(?:s)? occurred when fixing defaults:\n- At user\['some_section'\]\['some_list'\]:\s+Actual \(List\[float, float, int\]\) and declared \(List\[float\]\) types do not match\."
        ],
    ),
    (
        "input_errors",
        "input_type_error_str.yml",
        "template.yml",
        ParselglossyError,
        None,
        [
            r"Error(?:s)? occurred when fixing defaults:\n- At user\['some_section'\]\['a_short_string'\]:\s+Actual \(int\) and declared \(str\) types do not match\."
        ],
    ),
    (
        "input_errors",
        "input_predicate_intra.yml",
        "template.yml",
        ParselglossyError,
        None,
        [
            r"Error(?:s)? occurred when checking predicates:\n- At user\['some_section'\]\['another_number'\]:\s+Predicate 'value % 2 == 0' not satisfied\."
        ],
    ),
    (
        "input_errors",
        "input_predicate_cross.yml",
        "template.yml",
        ParselglossyError,
        None,
        [
            r"Error(?:s)? occurred when checking predicates:"
            r"\n- At user\['some_section'\]\['some_number'\]:\s+Predicate 'value < user\['some_section'\]\['another_number'\]' not satisfied\."
            r"\n- At user\['some_section'\]\['another_number'\]:\s+Predicate 'value > user\['some_section'\]\['some_number'\]' not satisfied\.",
            r"Error(?:s)? occurred when checking predicates:"
            r"\n- At user\['some_section'\]\['another_number'\]:\s+Predicate 'value > user\['some_section'\]\['some_number'\]' not satisfied\."
            r"\n- At user\['some_section'\]\['some_number'\]:\s+Predicate 'value < user\['some_section'\]\['another_number'\]' not satisfied\.",
        ],
    ),
]

validation_data = (
    [
        ("overall", "input.yml", "template.yml", None, valid(), [""]),
        (
            "overall",
            "default.yml",
            "template_with_default_section.yml",
            None,
            {"foobar": False, "foo": {"bar": False}},
            [""],
        ),
        (
            "overall",
            None,
            "template_all_default.yml",
            None,
            {"foobar": True, "foo": {"bar": False}},
            [""],
        ),
        (
            "overall",
            "input_ordering.yml",
            "ordering.yml",
            None,
            {
                "flavor": 2.0,
                "color": 2.0,
                "charge": 2.0,
                "spectator": "nothing to see",
                "adorable_kitten": {"flavor": 1.0, "color": 1.5, "charge": 1.0,},
            },
            [""],
        ),
    ]
    + template_errors_data
    + input_errors_data
)


def ids(terms: List[str]) -> str:
    return "-".join([t.rsplit(".", 1)[0] if t is not None else "" for t in terms])


@pytest.mark.parametrize(
    "folder,input_file_name,template_file_name,raises,valid,error_message",
    [
        pytest.param(f, i, t, r, v, e, id=ids([f, i, t]))
        for f, i, t, r, v, e in validation_data
    ],
)
def test_validation(
    folder, input_file_name, template_file_name, raises, valid, error_message
):
    user, template = read_in(folder, input_file_name, template_file_name)

    if raises is None:
        ctx = does_not_raise()
    else:
        ctx = pytest.raises(ParselglossyError, match="|".join(error_message))

    with ctx:
        # Validate and dump JSON
        dumped = Path("validated.json")
        template = is_template_valid(template)
        user = validate_from_dicts(ir=user, template=template, fr_file=dumped)
        assert user == valid

        # Read JSON (round-trip check)
        with dumped.open("r") as v:
            validated = json.load(v, object_hook=as_complex)
        assert validated == valid
        dumped.unlink()



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Getkw grammar generation."""

from typing import Any

from .atoms import (
    bool_t,
    complex_t,
    data_t,
    float_t,
    fortranStyleComment,
    int_t,
    make_list_t,
    quoted_str_t,
    unquoted_str_t,
)

try:
    import pyparsing as pp
except ImportError:
    # Import local copy
    from . import pyparsing as pp  # type: ignore


def grammar(*, has_complex: bool = False) -> Any:
    """The Getkw recursive grammar.

    Parameters
    ----------
    has_complex: bool
        Whether to include complex numbers. Defaults to `False`.

    Returns
    -------
    A parsing grammar.
    """

    EQ, COMMA = map(pp.Suppress, "=,")
    LBRACE, RBRACE = map(pp.Suppress, "{}")

    # Define key
    key = pp.Word(pp.alphas + "_<>", pp.alphanums + "_<>")

    # A scalar value (bool, int, float, str)
    if has_complex:
        scalar = quoted_str_t ^ complex_t ^ float_t ^ int_t ^ bool_t ^ unquoted_str_t
    else:
        scalar = quoted_str_t ^ float_t ^ int_t ^ bool_t ^ unquoted_str_t
    # Coerce lists to be lists
    list_t = make_list_t(scalar)
    list_t.setParseAction(lambda t: [t])

    # Define key-value pairs, i.e. our keywords
    pair = pp.Group(key + EQ + list_t) | pp.Group(key + EQ + scalar)

    # Define values and section recursively
    section = pp.Forward()
    values = pp.Forward()
    section << pp.Group(key + LBRACE + values + RBRACE)
    values << pp.Dict(pp.OneOrMore(section | data_t | pair))

    # Define input
    retval = pp.Dict(pp.OneOrMore(section) | pp.OneOrMore(values))

    # Ignore Python (#), C/C++ (/* */ and //), and Fortran (!) style comments
    comment = pp.cppStyleComment | pp.pythonStyleComment | fortranStyleComment
    retval.ignore(comment)

    return retval



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Tests for `parselglossy` package.

Tests merging of template and user input `dict`-s into an unvalidated input `dict`.
"""

from typing import Optional

import pytest

from parselglossy.exceptions import ParselglossyError
from parselglossy.validation import merge_ours
from parselglossy.views import view_by_default
from read_in import read_in

expected_data = [
    (
        "overall",
        "input.yml",
        "template.yml",
        {
            "scf": {
                "another_number": 10,
                "functional": "B3LYP",
                "max_num_iterations": 20,
                "some_acceleration": True,
                "some_complex_number": "0.0+0.0j",
                "thresholds": {"energy": 1.0e-5, "some_integral_screening": 0.0002},
            },
            "title": "this is an example",
        },
    ),
    (
        "overall",
        "default.yml",
        "template_with_default_section.yml",
        {"foobar": False, "foo": {"bar": False}},
    ),
    (
        "overall",
        None,
        "template_all_default.yml",
        {"foobar": True, "foo": {"bar": False}},
    ),
]


def expected_ids(name: Optional[str]) -> str:
    if name is None:
        id = "all_default"
    else:
        id = name.rsplit(".", 1)[0]
    return id


@pytest.mark.parametrize(
    "folder,user,template,ref",
    [pytest.param(*args, id=expected_ids(args[1])) for args in expected_data],
)
def test_merge_expected(folder, user, template, ref):
    user, template = read_in(folder, user, template)
    outgoing = merge_ours(theirs=view_by_default(template), ours=user)

    assert outgoing == ref


unexpected_data = [
    (
        "input_missing_keyword.yml",
        r"Error(?:s)? occurred when merging:\n- At user\['some_section'\]\['a_short_string'\]:\s+Keyword 'a_short_string' is required but has no value\.",
    ),
    (
        "unexpected_keyword.yml",
        r"Error(?:s)? occurred when merging:\n- Found unexpected keyword: 'strange'",
    ),
    (
        "unexpected_section.yml",
        r"Error(?:s)? occurred when merging:\n- Found unexpected section: 'weird'",
    ),
    (
        "unexpected_keyword_in_section.yml",
        r"Error(?:s)? occurred when merging:\n- Found unexpected keyword: 'strange'",
    ),
    (
        "unexpected_section_nested.yml",
        r"Error(?:s)? occurred when merging:\n- Found unexpected section: 'weird'",
    ),
]


@pytest.mark.parametrize(
    "input_file_name,error_message",
    [pytest.param(*args, id=expected_ids(args[0])) for args in unexpected_data],
)
def test_merge_unexpected(input_file_name, error_message):
    user, template = read_in("input_errors", input_file_name, "template.yml")

    with pytest.raises(ParselglossyError, match=error_message):
        outgoing = merge_ours(
            theirs=view_by_default(template), ours=user
        )  # type: Optional[JSONDict]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Tests for `parselglossy` package."""

import json
import re
import subprocess
from pathlib import Path
from shutil import rmtree

import pytest
from click.testing import CliRunner

from parselglossy import __version__, cli
from parselglossy.utils import as_complex


@pytest.mark.parametrize(
    "switch,expected",
    [
        ([], "Console script for parselglossy"),
        (["--version"], f"parselglossy, version {__version__}"),
        (["--help"], r"--help\s+Show this message and exit"),
    ],
    ids=["plain", "version", "help"],
)
def test_cli_switches(switch, expected):
    runner = CliRunner()
    result = runner.invoke(cli.cli, switch)
    assert result.exit_code == 0, f"{result.output}"
    assert re.search(expected, result.output, re.M) is not None


@pytest.mark.parametrize(
    "command,inp,references",
    [
        (
            [
                "generate",
                "--template",
                "tests/validation/overall/template.yml",
                "--target",
                "fufamu",
                "--grammar",
                "standard",
                "--docfile",
                "foo.rst",
            ],
            "tests/api/scf.inp",
            ("tests/ref/scf_fr.json", "tests/ref/generated_input.rst"),
        ),
        (
            [
                "generate",
                "--template",
                "tests/validation/overall/template.yml",
                "--target",
                "fufamu",
                "-g",
                f"{Path(__file__).parents[1].absolute() / 'parselglossy/grammars/atoms.py'}",
                "-g",
                f"{Path(__file__).parents[1].absolute() / 'parselglossy/grammars/getkw.py'}",
                "--tokenize",
                "from . import getkw; lexer = getkw.grammar(has_complex=True); ir = lexer.parseString(in_str).asDict()\n",
                "--docfile",
                "babar.rst",
            ],
            "tests/api/scf.inp",
            ("tests/ref/scf_fr.json", "tests/ref/generated_input.rst"),
        ),
    ],
    ids=["generate-default", "generated-custom-grammar"],
)
def test_cli_generate(command, inp, references):
    runner = CliRunner()
    result = runner.invoke(cli.cli, command)
    assert result.exit_code == 0, f"{result.output}"

    parser_dir = Path.cwd() / f"{command[4]}"

    # Check generated documentation agrees with reference
    ref_rst = Path(references[1]).resolve()
    dumped = parser_dir / f"docs/{command[-1]}"
    with ref_rst.open("r") as ref, dumped.open("r") as o:
        assert o.read() == ref.read()

    # write a front-end using the generated module
    front = Path("dwigt.py")
    with front.open("w") as buf:
        buf.write(f"from {command[4]} import cli\ncli.cli()")

    # spawn a subprocess to run the generated parser's CLI and parse an input file
    subprocess.run(["python", "dwigt.py", inp])

    # Check final representation matches with reference
    ref_json = Path(references[0]).resolve()
    out = Path("scf_fr.json").resolve()
    with ref_json.open("r") as ref, out.open("r") as o:
        assert json.loads(o.read(), object_hook=as_complex) == json.loads(
            ref.read(), object_hook=as_complex
        )

    # delete front-end and parsed output
    front.unlink()
    out.unlink()

    # clean up generated parser
    rmtree(parser_dir)



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Error-handling facilities."""

from collections import namedtuple
from typing import List

from .utils import location_in_dict


class ParselglossyError(Exception):
    """Exception raised when parsing fails."""

    pass


class Error(namedtuple("Error", ["address", "message"])):
    """Detailed error reporting for dictionaries.

    Attributes
    ----------
    address : Tuple
         The keys needed to access the offending element in the `dict`.
    message : str
         The error message.

    Notes
    -----
    As we need to support Python 3.6, we cannot use the defaults field of
    ``namedtuple`` or ``dataclass``.
    """

    __slots__ = ()

    def __new__(cls, address=(), message=""):
        return super(Error, cls).__new__(cls, address, message)

    def __eq__(self, other):
        return self.address == other.address and self.message == other.message

    def __repr__(self):
        msg = f"{self.message:s}"
        if self.address != ():
            msg = f"At {location_in_dict(address=self.address):s}:\n  {msg:s}"
        return "- " + msg


def collate_errors(*, when: str, errors: List[Error]) -> str:
    """Collate a list of error into an informative message.

    Parameters
    ----------
    when: str
        When the error occurred.
    errors: List[Error]
        List of errors.

    Returns
    -------
    msg: str
        An error message with details about where in the dictionary the error
        arose and what the error is. For example::

            Error(s) occurred when fixing defaults:
            - At 'user['scf']['another_number']'
                KeyError 'min_num_iterations' in closure
                'user['scf']['min_num_iterations'] / 2'
    """
    plural = "s" if len(errors) > 1 else ""
    preamble = f"\nError{plural:s} occurred when {when:s}:"
    msgs = [preamble] + [f"{e}" for e in errors]

    return "\n".join(msgs)



# -*- coding: utf-8 -*-
#
# parselglossy -- Generic input parsing library, speaking in tongues
# Copyright (C) 2020 Roberto Di Remigio, Radovan Bast, and contributors.
#
# This file is part of parselglossy.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
# For information on the complete list of contributors to the
# parselglossy library, see: <http://parselglossy.readthedocs.io/>
#

"""Tests for `parselglossy` package.

Tests type checking and type fixation.
"""

import pytest

from parselglossy.exceptions import ParselglossyError
from parselglossy.validation import fix_defaults


@pytest.fixture
def types():
    return {
        "title": "str",
        "some_float": "float",
        "some_section": {
            "a_short_string": "str",
            "some_number": "int",
            "another_number": "int",
            "some_feature": "bool",
            "some_list": "List[float]",
            "some_complex": "complex",
        },
    }


def type_error_bool():
    d = {
        "title": "this is my title",
        "some_float": 2.4,
        "some_section": {
            "a_short_string": "short",
            "some_number": 7,
            "another_number": 20,
            "some_feature": 1,
            "some_list": [1.0, 2.0, 3.0],
            "some_complex": complex(0.0, 1.0),
        },
    }

    n = "type_error_bool"
    e = r"""
Error occurred when fixing defaults:
- At user\['some_section'\]\['some_feature'\]:
  Actual \(int\) and declared \(bool\) types do not match\."""

    return d, n, e


def type_error_float():
    d = {
        "title": "this is my title",
        "some_float": 2,
        "some_section": {
            "a_short_string": "short",
            "some_number": 7,
            "another_number": 20,
            "some_feature": True,
            "some_list": [1.0, 2.0, 3.0],
            "some_complex": complex(0.0, 1.0),
        },
    }
    n = "type_error_float"
    e = r"""
Error occurred when fixing defaults:
- At user\['some_float'\]:
  Actual \(int\) and declared \(float\) types do not match\."""

    return d, n, e


def type_error_int():
    d = {
        "title": "this is my title",
        "some_float": 2.4,
        "some_section": {
            "a_short_string": "short",
            "some_number": 7.0,
            "another_number": 20,
            "some_feature": True,
            "some_list": [1.0, 2.0, 3.0],
            "some_complex": complex(0.0, 1.0),
        },
    }
    n = "type_error_int"
    e = r"""
Error occurred when fixing defaults:
- At user\['some_section'\]\['some_number'\]:
  Actual \(float\) and declared \(int\) types do not match\."""

    return d, n, e


def type_error_list():
    d = {
        "title": "this is my title",
        "some_float": 2.4,
        "some_section": {
            "a_short_string": "short",
            "some_number": 7,
            "another_number": 20,
            "some_feature": True,
            "some_list": [1.0, 2.0, 3],
            "some_complex": complex(0.0, 1.0),
        },
    }
    n = "type_error_list"
    e = r"""
Error occurred when fixing defaults:
- At user\['some_section'\]\['some_list'\]:
  Actual \(List\[float, float, int\]\) and declared \(List\[float\]\) types do not match\."""

    return d, n, e


def type_error_str():
    d = {
        "title": "this is my title",
        "some_float": 2.4,
        "some_section": {
            "a_short_string": 0,
            "some_number": 7,
            "another_number": 20,
            "some_feature": True,
            "some_list": [1.0, 2.0, 3.0],
            "some_complex": complex(0.0, 1.0),
        },
    }
    n = "type_error_str"
    e = r"""
Error(?:s)? occurred when fixing defaults:
- At user\['some_section'\]\['a_short_string'\]:
  Actual \(int\) and declared \(str\) types do not match\."""

    return d, n, e


type_checking_data = [
    type_error_bool(),
    type_error_float(),
    type_error_int(),
    type_error_list(),
    type_error_str(),
]


@pytest.mark.parametrize(
    "user,error_message", [pytest.param(d, e, id=n) for d, n, e in type_checking_data]
)
def test_input_errors(types, user, error_message):
    with pytest.raises(ParselglossyError, match=error_message):
        outgoing = fix_defaults(user, types=types)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['parselglossy', 'parselglossy.grammars']

package_data = \
{'': ['*']}

install_requires = \
['click', 'pyparsing', 'pyyaml', 'networkx']

extras_require = \
{'dev': ['black', 'flake8', 'isort', 'mypy'],
 'doc': ['sphinx', 'guzzle_sphinx_theme'],
 'test': ['codecov',
          'coverage',
          'hypothesis',
          'pytest',
          'pytest-black',
          'pytest-cov',
          'pytest-flake8',
          'pytest-mypy',
          'pytest-sugar']}

entry_points = \
{'console_scripts': ['parselglossy = parselglossy.cli:cli']}

setup(name='parselglossy',
      version='0.6.0',
      description='Top-level package for parselglossy.',
      author='Roberto Di Remigio',
      author_email=None,
      url='https://parselglossy.readthedocs.io',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      entry_points=entry_points,
      python_requires='~=3.6',
     )
